"""Methods common to binary files"""
import struct
import os
from collections import Counter

import numpy as np
import pyvista as pv

from pyansys._binary_reader import c_read_record
from pyansys import _binary_reader
from pyansys.errors import NoDistributedFiles

STRESS_TYPES = ['X', 'Y', 'Z', 'XY', 'YZ', 'XZ']
PRINCIPAL_STRESS_TYPES = ['S1', 'S2', 'S3', 'SINT', 'SEQV']
STRAIN_TYPES = ['X', 'Y', 'Z', 'XY', 'YZ', 'XZ', 'EQV']
THERMAL_STRAIN_TYPES = ['X', 'Y', 'Z', 'XY', 'YZ', 'XZ', 'EQV', 'ESWELL']

ANSYS_BINARY_FILE_TYPES = {2: 'Element Matrix File',
                           3: 'Element Save Data file',
                           4: 'Full stiffness-mass matrix File',
                           8: 'Substructure Matrices File',
                           9: 'Modal Results File',
                           10: 'Reduced Displacement File',
                           12: 'Result File',
                           16: 'Database File',
                           45: 'Component Mode Synthesis Matrices (CMS) File'}



# c *** standard usage of the block number (buffers) and file unit number(FUN)
# c     for the ansys files
# c
# c     File     FileName   Unit File    Block      Description
# c     EMAT ->    EMATNM       FUN02        2      element matrices
# c     ESAV ->    ESAVNM       FUN03        1      element save data
# c     R000 ->    R000NM       FUN03        7      element save data
# c     FULL ->    FULLNM       FUN04        3      full matrices
# c     EROT ->    EROTNM       FUN07        4      element matrices
# c     SUB  ->    SUBNM        FUN08        8/7    substructuring
# c     MODE ->    MODENM       FUN09        8 (7)  mode file
# c     RDSP ->    RDSPNM       FUN10        7      Reduced displacement file
# c     RFRQ ->    RFRQNM       FUN10        7      Reduced complex displacement (MSUP)
# c     TRI  ->    TRINM        FUN11        3      tri stiffness
# c     RST  ->    RSTNM        FUN12        6      results
# c     DSUB ->    DSUBNM       FUN13        5      displacement substructure
# c                             FUN14               input file?
# c                             FUN15               output file?
# c     RDB  ->                 FUN16               back up file
# c                             FUN17               temporary file (used in different cases)
# c     MODEL->    MODENM_LEFT  FUN19       19      mode file for left eigen modes
# c     SCR1 ->                 FUN22               scratch file
# c     SCR2 ->                 FUN23               scratch file
# c     SCR3 ->                 FUN24               scratch file
# c     SCR4 ->                 FUN26               scratch file
# c     TMP  ->                 FUN28               temporary file used for rewrite a file
# c     OPT  ->                 FUN36               optimization data file
# c     LN40 ->                 FUN40               Boeing reduced matrix
# c     LN41 ->                 FUN41               Boeing reduced matrix
# c     LN42 ->                 FUN42               Boeing reduced matrix
# c     MLV  ->                 FUN43        8      load vector file (substructure)
# c                             FUN44               ?
# c     IST  ->                 FUN63               initial stress file
# c     ASI  ->    ASIRSTNM     FUN66        9      asi results


class AnsysBinary():
    """ANSYS binary file class"""
    filename = None

    def read_record(self, pointer, return_bufsize=False):
        """Reads a record at a given position.

        Because ANSYS 19.0+ uses compression by default, you must use
        this method rather than ``np.fromfile``.

        Parameters
        ----------
        pointer : int
            ANSYS file position (n words from start of file).  A word
            is four bytes.

        return_bufsize : bool, optional
            Returns the number of words read (includes header and
            footer).  Useful for determining the new position in the
            file after reading a record.

        Returns
        -------
        record : np.ndarray
            The record read as a ``n x 1`` numpy array.

        bufsize : float, optional
            When ``return_bufsize`` is enabled, returns the number of
            words read.

        """
        return c_read_record(self.filename, pointer, return_bufsize)


def read_binary(filename, **kwargs):
    """Reads ANSYS-written binary files:
    - Jobname.RST: Result file from structural analysis
    - Jobname.RTH: Result file from a thermal analysis
    - Jobname.EMAT: Stores data related to element matrices
    - Jobname.FULL: Stores the full stiffness-mass matrix

    Parameters
    ----------
    filename : str
        Filename to read.

    **kwargs : keyword arguments
        See the individual classes for additional keyword arguments.

    Examples
    --------
    >>> import pyansys
    >>> result = pyansys.read_binary('file.rst')
    >>> result = pyansys.read_binary('file.rst')
    >>> full_file = pyansys.read_binary('file.full')
    >>> emat_file = pyansys.read_binary('file.emat')

    Notes
    -----
    The following file types are unsupported
    - Jobname.DSUB file, storing displacements related to substructure
      matrices
    - Jobname.SUB file, storing data related to substructure matrices
    - Jobname.RFRQ file, storing data related to a mode-superposition
      harmonic analysis
    - The Jobname.RDSP file, storing data related to a
      mode-superposition transient analysis.
    - Jobname.MODE file, storing data related to a modal analysis
    - Jobname.RMG A magnetic analysis
    - Jobname.RFL A FLOTRAN analysis (a legacy results file)
    """
    if not os.path.isfile(filename):
        raise FileNotFoundError('%s is not a file or cannot be found' %
                                str(filename))

    file_format = read_standard_header(filename)['file format']

    if file_format == 2:
        from pyansys.emat import EmatFile
        return EmatFile(filename, **kwargs)
    elif file_format == 4:
        from pyansys.full import FullFile
        return FullFile(filename, **kwargs)
    elif file_format == 12:
        from pyansys.rst import Result
        from pyansys.dis_result import DistributedResult
        read_mesh = kwargs.pop('read_mesh', True)
        result = Result(filename, read_mesh=False, **kwargs)

        if result._is_distributed:
            try:  # can't find any files!
                return DistributedResult(filename, **kwargs)
            except NoDistributedFiles:
                # simply try to treat it as a non-distributed file
                pass

        # check if it's a cyclic result file
        ignore_cyclic = kwargs.pop('ignore_cyclic', False)
        if result._is_cyclic and not ignore_cyclic:
            from pyansys.cyclic_reader import CyclicResult
            return CyclicResult(filename)

        if read_mesh:
            result._store_mesh()

        return result

    # elif file_format == 16:
    #     from pyansys.db import Database
    #     return Database(filename, debug=kwargs.pop('debug', False))

    # No file matches
    file_type = ANSYS_BINARY_FILE_TYPES.get(file_format, str(file_format))
    raise RuntimeError('ANSYS binary "%s" not supported' % file_type)


def read_table(f, dtype='i', nread=None, skip=False, get_nread=True, cython=False):
    """ read fortran style table """
    if cython:
        arr, bufsz = c_read_record(f.name, f.tell()//4, True)
        f.seek(bufsz*4, 1)
        return arr

    if get_nread:
        n = np.fromfile(f, 'i', 1)
        if not n:
            raise Exception('end of file')

        tablesize = n[0]
        if dtype is None:  # read flags to get data type
            # ansys_dtype = np.fromfile(f, 'i', 1)
            flags = f.read(4)[-1]
            type_flag = flags >> 0 & 1
            prec_flag = flags >> 1 & 1
            # zlib_flag = flags >> 2 & 1
            sparse_flag = flags >> 3 & 1
            if sparse_flag:
                raise NotImplementedError('Cannot read sparse results.\nPlease run with "/FCOMP,RST,0" to disable writing sparse results')

            if type_flag:
                if prec_flag:
                    dtype = 'short'
                else:
                    dtype = 'i'
            else:
                if prec_flag:
                    dtype = 'float'
                else:
                    dtype = 'double'
        else:
            f.seek(4, 1)  # skip padding

    # override
    if nread:
        tablesize = nread

    if skip:
        f.seek((tablesize + 1)*4, 1)
        return tablesize
    else:
        if dtype == 'double':
            tablesize //= 2
        table = np.fromfile(f, dtype, tablesize)
    f.seek(4, 1)  # skip padding
    return table


def read_string_from_binary(f, n):
    """Read n 4 character binary strings from a file opend in binary
    mode
    """
    string = b''
    for _ in range(n):
        string += f.read(4)[::-1]

    try:
        return string.decode('utf')
    except:
        return string


def parse_header(table, keys):
    """ parses a header from a table """
    header = {}
    max_entry = len(table) - 1
    # some keys occure multiple times and refer to arrays of some sort
    counter = Counter(keys)
    del counter['0']

    # initialize lists/arrays
    for key, count in counter.items():
        if count > 1:
            header[key] = []

    for i, key in enumerate(keys):
        if i > max_entry:
            header[key] = 0
        else:
            if counter[key]>1:  # multiple items in the header -> list/array
                if table[i]:  # skip zeros
                    header[key].append(table[i])
            else:
                header[key] = table[i] 

    for key in keys:
        if 'ptr' in key and key[-1] == 'h':
            basekey = key[:-1]
            intl = header[basekey + 'l']
            inth = header[basekey + 'h']
            header[basekey] = two_ints_to_long(intl, inth)

    # remove empty entries
    header.pop('_', None)
    return header


def two_ints_to_long(intl, inth):
    """ Interpert two ints as one long """
    longint = struct.pack(">I", inth) + struct.pack(">I", intl)
    return struct.unpack('>q', longint)[0]


def read_standard_header(filename):
    """ Reads standard header """
    with open(filename, 'rb') as f:

        endian = '<'
        if np.fromfile(f, dtype='<i', count=1) != 100:

            # Check if big enos
            f.seek(0)
            if np.fromfile(f, dtype='>i', count=1) == 100:
                endian = '>'

            # Otherwise, it's probably not a result file
            else:
                raise RuntimeError('Unable to determine endian type.  ' +
                                   'Possibly not an ANSYS binary file')

        f.seek(0)

        header = {}
        header['endian'] = endian
        header['file number'] = read_table(f, nread=1, get_nread=False)[0]
        header['file format'] = read_table(f, nread=1, get_nread=False)[0]
        int_time = str(read_table(f, nread=1, get_nread=False)[0])
        header['time'] = ':'.join([int_time[0:2], int_time[2:4], int_time[4:]])
        int_date = str(read_table(f, nread=1, get_nread=False)[0])
        if int_date == '-1':
            header['date'] = ''
        else:
            header['date'] = '/'.join([int_date[0:4], int_date[4:6], int_date[6:]])

        unit_types = {0: 'User Defined',
                      1: 'SI',
                      2: 'CSG',
                      3: 'U.S. Customary units (feet)',
                      4: 'U.S. Customary units (inches)',
                      5: 'MKS',
                      6: 'MPA',
                      7: 'uMKS'}
        header['units'] = unit_types[read_table(f, nread=1, get_nread=False)[0]]

        f.seek(11 * 4)
        version = read_string_from_binary(f, 1).strip()

        header['verstring'] = version
        header['mainver'] = int(version[:2])
        header['subver'] = int(version[-1])

        # there's something hidden at 12
        f.seek(4, 1)

        # f.seek(13 * 4)
        header['machine'] = read_string_from_binary(f, 3).strip()
        header['jobname'] = read_string_from_binary(f, 2).strip()
        header['product'] = read_string_from_binary(f, 2).strip()
        header['special'] = read_string_from_binary(f, 1).strip()
        header['username'] = read_string_from_binary(f, 3).strip()

        # Items 23-25 The machine identifier in integer form (three four-character strings)
        # this contains license information
        header['machine_identifier'] = read_string_from_binary(f, 3).strip()

        # Item 26 The system record size
        header['system record size'] = read_table(f, nread=1, get_nread=False)[0]

        # Item 27 The maximum file length
        # header['file length'] = read_table(f, nread=1, get_nread=False)[0]

        # Item 28 The maximum record number
        # header['the maximum record number'] = read_table(f, nread=1, get_nread=False)[0]

        # Items 31-38 The Jobname (eight four-character strings)
        f.seek(32*4)
        header['jobname2'] = read_string_from_binary(f, 8).strip()

        # Items 41-60 The main analysis title in integer form (20 four-character strings)
        f.seek(42*4)
        header['title'] = read_string_from_binary(f, 20).strip()

        # Items 61-80 The first subtitle in integer form (20 four-character strings)
        header['subtitle'] = read_string_from_binary(f, 20).strip()

        # Item 95 The split point of the file (0 means the file will not split)
        f.seek(96*4)
        header['split point'] = read_table(f, nread=1, get_nread=False)[0]

        # Items 97-98 LONGINT of the maximum file length (bug here)
        # ints = read_table(f, nread=2, get_nread=False)
        # header['file length'] = two_ints_to_long(ints[0], ints[1])

    return header


def rotate_to_global(result, euler_angles):
    """Rotate a result set to the global coordinate system

    ANSYS writes the results in the nodal coordinate system and
    they use the ZXY euler rotation method.

    Rotates results in-place.

    Parameters
    ----------
    result : np.ndarray
        Nodal result set to rotate.  Sized ``n_node`` x ``ndof``

    euler_angles : np.ndarray
        Array of euler angles sized 3 x ``n_node`` in degrees.

    """
    theta_xy, theta_yz, theta_zx = euler_angles
    if np.any(theta_xy):
        pv.common.axis_rotation(result, theta_xy, inplace=True, axis='z')

    if np.any(theta_yz):
        pv.common.axis_rotation(result, theta_yz, inplace=True, axis='x')

    if np.any(theta_zx):
        pv.common.axis_rotation(result, theta_zx, inplace=True, axis='y')



# cython: embedsignature=True
# cython: language_level=3
# cython: boundscheck=False
# cython: wraparound=False
# cython: cdivision=True

import ctypes
import numpy as np

from libc.math cimport sqrt, fabs, sin, cos
from libc.stdio cimport (fopen, FILE, fclose, fread, fseek, SEEK_CUR,
                         ftell, SEEK_SET)
from libc.string cimport memcpy
from libc.stdint cimport int64_t, int32_t
from libc.stdlib cimport malloc, free

# debug
from libc.stdio cimport printf


from cython.parallel import prange
ctypedef unsigned char uint8


cdef extern from "<iostream>" namespace "std" nogil:
     cdef cppclass ostream:
          ostream& write(const char*, int) except +
     cdef cppclass istream:
          istream& read(const char*, int) except +
     cdef cppclass ifstream(istream):
          ifstream(const char *, open_mode) except +


cdef extern from "<fstream>" namespace "std" nogil:          
     cdef cppclass filebuf:
          pass
        
     cdef cppclass fstream:
          void close()
          bint is_open()
          void open(const char*, open_mode)
          void open(const char&, open_mode)
          filebuf* rdbuf() const
          filebuf* rdbuf(filebuf* sb)

     cdef cppclass ifstream(istream):
          # void close()
          ifstream(const char*) except +
          ifstream(const char*, open_mode) except+

cdef extern from "<iostream>" namespace "std::ios_base" nogil:
     cdef cppclass open_mode:
          pass
     cdef open_mode binary


cdef extern from "numpy/npy_math.h" nogil:
    bint npy_isnan(double x)

cdef extern from 'binary_reader.h' nogil:
    void read_nodes(const char*, int64_t, int, int*, double*)
    void* read_record(const char*, int64_t, int*, int*, int*, int*)
    void read_record_stream(ifstream*, int64_t, void*, int*, int*, int*)


# VTK numbering for vtk cells
cdef uint8 VTK_EMPTY_CELL = 0
cdef uint8 VTK_VERTEX = 1
cdef uint8 VTK_LINE = 3

cdef uint8 VTK_TRIANGLE = 5
cdef uint8 VTK_QUAD = 9
cdef uint8 VTK_QUADRATIC_TRIANGLE = 22
cdef uint8 VTK_QUADRATIC_QUAD = 23

cdef uint8 VTK_TETRA = 10
cdef uint8 VTK_HEXAHEDRON = 12
cdef uint8 VTK_WEDGE = 13
cdef uint8 VTK_PYRAMID = 14
cdef uint8 VTK_QUADRATIC_TETRA = 24
cdef uint8 VTK_QUADRATIC_PYRAMID = 27
cdef uint8 VTK_QUADRATIC_WEDGE = 26
cdef uint8 VTK_QUADRATIC_HEXAHEDRON = 25

cdef double DEG2RAD = 0.017453292519943295

ctypedef fused index_type:
    int
    int64_t

ctypedef fused float_or_double:
    float
    double


###############################################################################
from libc.stdlib cimport free
from cpython cimport PyObject, Py_INCREF

import numpy as np  # Python-level symbols of numpy
cimport numpy as np  # C-level symbols of numpy

# Numpy must be initialized. When using numpy from C or Cython you must
# _always_ do that, or you will have segfaults
np.import_array()


cdef class ArrayWrapper:
    """Array wrapper class to deallocate our array when the Python
    object is deleted."""
    cdef void* data_ptr
    cdef int size
    cdef int my_dtype

    cdef set_data(self, int size, void* data_ptr, int my_dtype=0):
        """Set the data of the array.

        This cannot be done in the constructor as it must recieve C-level
        arguments.

        Parameters
        ----------
        size : int
            Length of the array

        data_ptr : void*
            Pointer to the data            
        """
        self.data_ptr = data_ptr
        self.size = size        
        self.my_dtype = my_dtype

    def __array__(self):
        """ Here we use the __array__ method, that is called when numpy
            tries to get an array from the object."""
        cdef np.npy_intp shape[1]
        shape[0] = <np.npy_intp> self.size
        # Create a 1D array, of length 'size'
        if self.my_dtype == 0:
            ndarray = np.PyArray_SimpleNewFromData(1, shape,
                                                   np.NPY_INT16, self.data_ptr)
        elif self.my_dtype == 1:
            ndarray = np.PyArray_SimpleNewFromData(1, shape,
                                                   np.NPY_INT32, self.data_ptr)
        elif self.my_dtype == 2:
            ndarray = np.PyArray_SimpleNewFromData(1, shape,
                                                   np.NPY_FLOAT32, self.data_ptr)
        else:
            ndarray = np.PyArray_SimpleNewFromData(1, shape,
                                                   np.NPY_FLOAT64, self.data_ptr)

        return ndarray

    def __dealloc__(self):
        """ Frees the array. This is called by Python when all the
        references to the object are gone. """
        free(<void*>self.data_ptr)


###############################################################################
# element result location pointers
cdef int PTR_ENS_IDX = 2
cdef int PTR_EEL_IDX = 5
cdef int PTR_EUL_IDX = 9


cdef inline double get_double(char * array) nogil:
    cdef double result
    memcpy(&result, array, sizeof(result))
    return result


cdef inline double get_float(char * array) nogil:
    cdef float result
    memcpy(&result, array, sizeof(result))
    return result


cdef inline int get_int(char * array) nogil:
    cdef int result
    memcpy(&result, array, sizeof(result))
    return result


def load_nodes(filename, int ptr_loc, int nnod, double [:, ::1] nloc, 
              int [::1] nnum):
    """Wrapper for cpp function

    """    
    cdef bytes buf, flags_buf
    cdef bytes py_bytes = filename.encode()
    cdef char* c_filename = py_bytes
    read_nodes(c_filename, ptr_loc, nnod, &nnum[0], &nloc[0, 0])


def c_read_record(filename, int64_t ptr, int return_bufsize=0):
    """Read an ANSYS record and return a numpy array"""
    cdef bytes py_bytes = filename.encode()
    cdef char* c_filename = py_bytes

    cdef int prec_flag, type_flag, size, my_dtype, bufsize
    cdef void* c_ptr
    c_ptr = read_record(c_filename, ptr, &prec_flag, &type_flag, &size, &bufsize)
    cdef np.ndarray ndarray = wrap_array(c_ptr, size, type_flag, prec_flag)

    if return_bufsize:
        return ndarray, bufsize
    else:
        return ndarray


cdef np.ndarray wrap_array(void* c_ptr, int size, int type_flag, int prec_flag):
    """wrap a c array as a numpy array"""
    cdef int my_dtype

    if type_flag:
        if prec_flag:
            my_dtype = 0  # np.NPY_INT16
        else:
            my_dtype = 1  # np.NPY_INT32
    else:
        if prec_flag:
            my_dtype = 2  # np.NPY_FLOAT32
        else:
            my_dtype = 3  # np.NPY_FLOAT64

    # wrap c_array 
    array_wrapper = ArrayWrapper()
    array_wrapper.set_data(size, c_ptr, my_dtype)

    cdef np.ndarray ndarray
    ndarray = np.array(array_wrapper, copy=False)

    # Assign our object to the 'base' of the ndarray object
    ndarray.base = <PyObject*> array_wrapper

    # Increment the reference count, as the above assignement was done in
    # C, and Python does not know that there is this additional reference
    Py_INCREF(array_wrapper)

    return ndarray


def load_elements(filename, int64_t loc, int nelem, int64_t [::1] e_disp_table):
    """Load elements from an ansys result file.

    filename : str
        Filename containing the result file.

    loc : int64_t
        Pointer to the element table

    e_disp_table : int64_t [::1]
        Array containing pointers to the individual elements relative to ``loc``.

    The following is stored for each element
    0 - mat     - material reference number
    1 - type    - element type number
    2 - real    - real constant reference number
    3 - secnum  - section number
    4 - esys    - element coordinate system
    5 - death   - death flat (1 live, 0 dead)
    6 - solidm  - solid model reference
    7 - shape   - coded shape key
    8 - elnum   - element number
    9 - baseeid - base element number
    10 - NODES   - node numbers defining the element
    """
    cdef int i, j

    cdef int prec_flag, type_flag, size, bufsize
    cdef void* c_ptr

    cdef bytes py_bytes = filename.encode()
    cdef char* c_filename = py_bytes
    cdef int* element
    cdef short* s_element

    cdef int val, nread
    cdef int64_t elem_loc

    # elem connectivity and info (10 fields + maximum of 20 nodes per element)
    cdef int [::1] elem = np.empty(nelem*30, np.int32)
    cdef int [::1] elem_off = np.empty(nelem + 1, np.int32)

    cdef int c = 0  # cell position counter
    for i in range(nelem):
        # load element
        elem_loc = loc + e_disp_table[i]
        c_ptr = read_record(c_filename, elem_loc, &prec_flag, &type_flag,
                            &size, &bufsize)

        # start of the element
        elem_off[i] = c

        # read in entire element
        if prec_flag:
            s_element = <short*>c_ptr
            for j in range(size):
                elem[c + j] = s_element[j]
        else:
            element = <int*>c_ptr
            for j in range(size):
                elem[c + j] = element[j]
        c += size

    # add final position here for parser to know the size of the last element
    elem_off[nelem] = c

    return np.array(elem[:c]), np.array(elem_off)


def read_element_stress(filename, int64_t [::1] ele_ind_table, 
                        int64_t [::1] nodstr, int [::1] etype,
                        double [:, ::1] ele_data_arr, int nitem,
                        int [::1] element_type, int64_t ptr_off,
                        int as_global=1):
    """Read element results from ANSYS directly into a numpy array

    ele_ind_table : int64_t [::1]
        Pointer to the result header of an element relative to the
        result index.

    ptr_off : int
        ``ele_ind_table`` offset from the file head.

    as_global : int, optional
        Rotates stresses from the element coordinate system to the global
        cartesian coordinate system.  Default True.
    """
    cdef bytes py_bytes = filename.encode()
    cdef char* c_filename = py_bytes
    cdef ifstream* binfile = new ifstream(c_filename, binary)

    cdef int i
    cdef int c = 0
    for i in range(len(ele_ind_table)):
        nnode_elem = nodstr[etype[i]]

        if ele_ind_table[i] != 0:
            read_element_result(binfile, ele_ind_table[i] + ptr_off,
                                PTR_ENS_IDX, nnode_elem, nitem,
                                &ele_data_arr[c, 0], as_global)
            c += nnode_elem


def populate_surface_element_result(filename,
                                    int64_t [::1] ele_ind_table, 
                                    int [::1] nodstr,
                                    int [::1] etype,
                                    int nitem,
                                    int64_t ptr_off,
                                    int elem_result_index,
                                    int n_points,
                                    int64_t [::1] faces,
                                    int n_faces,
                                    int [::1] nnum_surf,
                                    int64_t [::1] elem_ind,
                                    int [::1] elem,
                                    int [::1] elem_off,
                                    int item_index,
                                    int as_global=1):
    """Read element results from ANSYS directly into a numpy array

    filename : string
        Filename of the result file.

    ele_ind_table :  [::1] int64_t
        Pointer to the result header of an element relative to the
        result index.

    nodstr : [::1] int
        Number of elements per element.  Given on an element by element basis.

    etype : [::1] int
        Array containing the element type of each element.

    nitem : int
        Number of data items per node.

    ptr_off : int
        ``ele_ind_table`` offset from the file head.

    elem_result_index  : int
        Index within the element table pointing to the data of interest

    n_points : int
        Number of points in the surface.

    faces : [::1] int64_t
        VTK style face array containing the number of points per face
        folled by the face connectivity.

    n_faces : int
        Number of faces in the surface.

    nnum_surf : [::1] int
        Node number of the point within a face.  Similar format to
        ``faces`` but without the indices indicating the size of the
        face.

    elem_ind : [::1] int64_t
        Index of the element within elem for each face.

    elem : [::1] int
        Element data array containing element type and connectivity.

    elem_off : [::1] int
        Array of offsets to the start of each individual element in ``elem``.

    item_index : int
        Index of the data item for each node within the element.

    as_global : int, optional
        Rotates stresses from the element coordinate system to the global
        cartesian coordinate system.  Default True.

    Returns
    -------
    data : [::1] double
        Array of data to output.  One value of data for each point.


    """
    # open the result file
    cdef bytes py_bytes = filename.encode()
    cdef char* c_filename = py_bytes
    cdef ifstream* binfile = new ifstream(c_filename, binary)

    # temp buffer for the element data
    # best to keep this large as nitem might be wrong
    cdef double [::1] data_buff = np.empty(4096)  # nitem*number_of_nodes
    cdef double [::1] data = np.empty(n_points)

    cdef int i, j, k, fsize, st, node_num, elem_nnum, nnode_elem
    cdef int64_t elem_idx
    cdef int c = 0   # includes face size increment
    cdef int cj = 0   # does not include face size increment

    # loop through each face
    for i in range(n_faces):
        elem_idx = elem_ind[i]  # index of the element this face references
        data_off = ele_ind_table[elem_idx]  # offset from ptr_off
        nnode_elem = nodstr[etype[elem_idx]]

        # number of points in the face
        fsize = faces[c]
        c += fsize + 1

        # read the element if the data exists
        if data_off != 0:
            read_element_result(binfile, data_off + ptr_off,
                                elem_result_index, nnode_elem, nitem,
                                &data_buff[0], as_global)

            # start of the node numbers within the element data
            st = elem_off[elem_idx] + 10

            # populate dat array with the nodal data
            for j in range(fsize):
                # node number of this point on the face
                node_num = nnum_surf[cj]

                # loop through each node in the element until a match is found
                for k in range(nnode_elem):
                    elem_nnum = elem[st + k]
                    if elem_nnum == node_num:
                        data[cj] = data_buff[k*nitem + item_index]
                        break

                cj += 1

        else:  # populate with zeros (though maybe NAN is better?)
            for j in range(fsize):
                data[cj] = 0
                cj += 1

    return np.array(data)


cdef inline int read_element_result(ifstream *binfile, int64_t ele_table,
                                    int result_index,
                                    int nnode_elem, int nitem, double *arr,
                                    int as_global=1):
    """Populate array with results from a single element"""
    cdef int i, j, k, c, nitems
    cdef int [4096] pointers  # tmp array of pointers
    cdef int prec_flag, type_flag, size
    cdef int64_t ptr, eul_ptr
    cdef char [131072] tmp_data_buffer  # 2**17
    cdef double [512] euler_angles  # 8*3*20 --> 512
    cdef short* spointers = <short*>pointers

    # store elemenet element result pointers
    read_record_stream(binfile, ele_table, <void*>&pointers,
                       &prec_flag, &type_flag, &size)
    # expect size to be 25 here as of v19.1

    # always cast 
    if prec_flag:
        ptr = spointers[result_index]
        eul_ptr = spointers[PTR_EUL_IDX]
    else:
        ptr = pointers[result_index]
        eul_ptr = pointers[PTR_EUL_IDX]

    if ptr == 0:  # 0 means skip
        return 1
    if ptr < 0:  # negative pointer means missing data
        # skip this element
        for j in range(nnode_elem):
            for k in range(nitem):
                arr[j*nitem + k] = 0  # consider putting NAN instead
    else:
        # read the results evaluated at the integration points or nodes
        read_record_stream(binfile, ele_table + ptr, <void*>tmp_data_buffer,
                           &prec_flag, &type_flag, &size)

        # if size < 0:
        #     raise MemoryError('Negative array size')

        # copy to main array
        if prec_flag:
            for i in range(size):
                arr[i] = <double>(<float*>tmp_data_buffer)[i]
        else:
            # EFFICIENCY WARNING: we don't need to copy here
            for i in range(size):
                arr[i] = (<double*>tmp_data_buffer)[i]

        # rotate out of element coordinate system
        if as_global and eul_ptr > 0 and result_index == PTR_ENS_IDX:  # or PTR_EEL_IDX
            # read in euler angles
            read_record_stream(binfile, ele_table + eul_ptr,
                               <void*>tmp_data_buffer, &prec_flag, &type_flag, &size)

            # copy to main array
            if prec_flag:
                for i in range(size):
                    euler_angles[i] = <double>(<float*>tmp_data_buffer)[i]
            else:  # we don't need to copy here...
                for i in range(size):
                    euler_angles[i] = (<double*>tmp_data_buffer)[i]            

            if size == 3:
                # --For uniform reduced integration lower-order 
                # elements (e.g. PLANE182, KEYOPT(1)=1 and
                # SOLID185 KEYOPT(2)=1):
                # the angles are at the centroid and the number
                # of items is 3.
                # if element_type == 181 or element_type == 281:
                    # euler_rotate_shell(arr, euler_angles, nitem)
                # else:
                euler_rotate(arr, euler_angles, nitem, nnode_elem)
            else:
                for i in range(nnode_elem):
                    euler_rotate(&arr[i*nitem], &euler_angles[3*i], nitem, 1)
                # --For other formulations of lower-order 
                # elements (e.g. PLANE182 and SOLID185) and
                # the higher-order elements
                # (e.g. PLANE183, SOLID186, and SOLID187):
                # The number of items in this record is
                # (nodstr*3).

            # TODO: NOT IMPLEMENTED
            # --For layered solid elements, add NL values,
            # so that the number of items in this record 
            # is (nodstr*3)+NL.

    return 0


cdef inline void euler_rotate_shell(float_or_double *arr,
                                    float_or_double [64] eulerangles, int nitem) nogil:
    """Performs a 3-1-2 euler rotation given thxy, thyz, thzx in
    ``eulerangles``

    Acts on rows 0 - 3 relative to row

    Specific to shell181 elements

    # used sympy to generate these equations
    tensor = np.matrix([[s_xx, s_xy, s_xz], 
                        [s_xy, s_yy, s_yz], 
                        [s_xz, s_yz, s_zz]])

    # always zero for shell elements...
    s_xz = 0
    s_yz = 0
    s_zz = 0

    from sympy import Matrix, symbols

    c1, c2, c3, s1, s2, s3, s_xx, s_yy, s_xy = symbols('c1 c2 c3 s1 s2 s3 s_xx s_yy s_xy')
    tensor = np.matrix([[s_xx, s_xy, 0], [s_xy, s_yy, 0], [0, 0, 0]])
    

    R = Matrix([[c1*c3 - s1*s2*s3, s1*c3 + c1*s2*s3, -s3*c2],
                [-s1*c2, c1*c2, s2],
                [c1*s3 + s1*s2*c3, s1*s3 - c1*c3*s2, c2*c3]])

    ans = R.T*tensor*R
    """    
    cdef double s_xx, s_xy, s_yy
    cdef double c1 = cos(DEG2RAD*eulerangles[0])
    cdef double c2 = cos(DEG2RAD*eulerangles[1])
    cdef double c3 = cos(DEG2RAD*eulerangles[2])
    cdef double s1 = sin(DEG2RAD*eulerangles[0])
    cdef double s2 = sin(DEG2RAD*eulerangles[1])
    cdef double s3 = sin(DEG2RAD*eulerangles[2])

    cdef int i
    for i in range(4):
        # grab the element component stresses
        s_xx = arr[i*nitem + 0]
        s_yy = arr[i*nitem + 1]
        s_xy = arr[i*nitem + 3]
        # the rest are zero (no out of plane stress)

        arr[i*nitem + 0] = -c2*s1*(-c2*s1*s_yy + s_xy*(c1*c3 - s1*s2*s3)) + (c1*c3 - s1*s2*s3)*(-c2*s1*s_xy + s_xx*(c1*c3 - s1*s2*s3))
        arr[i*nitem + 1] = c1*c2*(c1*c2*s_yy + s_xy*(c1*s2*s3 + c3*s1)) + (c1*c2*s_xy + s_xx*(c1*s2*s3 + c3*s1))*(c1*s2*s3 + c3*s1)
        arr[i*nitem + 2] = -c2*s3*(-c2*s3*s_xx + s2*s_xy) + s2*(-c2*s3*s_xy + s2*s_yy)
        arr[i*nitem + 3] = c1*c2*(-c2*s1*s_yy + s_xy*(c1*c3 - s1*s2*s3)) + (c1*s2*s3 + c3*s1)*(-c2*s1*s_xy + s_xx*(c1*c3 - s1*s2*s3))
        arr[i*nitem + 4] = -c2*s3*(c1*c2*s_xy + s_xx*(c1*s2*s3 + c3*s1)) + s2*(c1*c2*s_yy + s_xy*(c1*s2*s3 + c3*s1))
        arr[i*nitem + 5] = -c2*s3*(-c2*s1*s_xy + s_xx*(c1*c3 - s1*s2*s3)) + s2*(-c2*s1*s_yy + s_xy*(c1*c3 - s1*s2*s3))


cdef inline void euler_rotate(float_or_double *arr,
                              float_or_double [64] eulerangles, int nitem,
                              int n_node):
    """Performs a 3-1-2 euler rotation given thxy, thyz, thzx in
    ``eulerangles`` on the stress values in ``arr``

    Notes
    -----
    Used sympy to generate these equations
    import numpy as np
    from sympy import Matrix, symbols
    c1, c2, c3, s1, s2, s3 = symbols('c1 c2 c3 s1 s2 s3')
    s_xx, s_xy, s_yy, s_xz, s_yz, s_zz = symbols('s_xx s_xy s_yy s_xz s_yz s_zz')

    tensor = np.matrix([[s_xx, s_xy, s_xz],
                        [s_xy, s_yy, s_yz],
                        [s_xz, s_yz, s_zz]])

    # from: https://www.simutechgroup.com/tips-and-tricks/fea-articles/286-fea-tips-tricks-ansys-rotation-convention
    R = Matrix([[c1*c3 - s1*s2*s3, s1*c3 + c1*s2*s3, -s3*c2],
                [-s1*c2, c1*c2, s2],
                [c1*s3 + s1*s2*c3, s1*s3 - c1*c3*s2, c2*c3]])

    ans = R.T*tensor*R


    print('XX', ans[0, 0])
    print('YY', ans[1, 1])
    print('ZZ', ans[2, 2])

    print('XY', ans[0, 1])
    print('YZ', ans[1, 2])
    print('XZ', ans[0, 2])
    """    
    cdef double s_xx, s_xy, s_yy, s_xz, s_yz, s_zz
    cdef double c1 = cos(DEG2RAD*eulerangles[0])
    cdef double c2 = cos(DEG2RAD*eulerangles[1])
    cdef double c3 = cos(DEG2RAD*eulerangles[2])
    cdef double s1 = sin(DEG2RAD*eulerangles[0])
    cdef double s2 = sin(DEG2RAD*eulerangles[1])
    cdef double s3 = sin(DEG2RAD*eulerangles[2])

    # rotate each node in the element
    cdef int i
    for i in range(n_node):
        # grab the node component stresses
        s_xx = arr[i*nitem + 0]
        s_yy = arr[i*nitem + 1]
        s_zz = arr[i*nitem + 2]
        s_xy = arr[i*nitem + 3]
        s_yz = arr[i*nitem + 4]
        s_xz = arr[i*nitem + 5]

        # store rotated component stresses 
        # XX
        arr[i*nitem + 0] = -c2*s1*(-c2*s1*s_yy + s_xy*(c1*c3 - s1*s2*s3) + s_yz*(c1*s3 + c3*s1*s2)) + (c1*c3 - s1*s2*s3)*(-c2*s1*s_xy + s_xx*(c1*c3 - s1*s2*s3) + s_xz*(c1*s3 + c3*s1*s2)) + (c1*s3 + c3*s1*s2)*(-c2*s1*s_yz + s_xz*(c1*c3 - s1*s2*s3) + s_zz*(c1*s3 + c3*s1*s2))

        # YY
        arr[i*nitem + 1] = c1*c2*(c1*c2*s_yy + s_xy*(c1*s2*s3 + c3*s1) + s_yz*(-c1*c3*s2 + s1*s3)) + (-c1*c3*s2 + s1*s3)*(c1*c2*s_yz + s_xz*(c1*s2*s3 + c3*s1) + s_zz*(-c1*c3*s2 + s1*s3)) + (c1*s2*s3 + c3*s1)*(c1*c2*s_xy + s_xx*(c1*s2*s3 + c3*s1) + s_xz*(-c1*c3*s2 + s1*s3))

        # ZZ
        arr[i*nitem + 2] = c2*c3*(c2*c3*s_zz - c2*s3*s_xz + s2*s_yz) - c2*s3*(c2*c3*s_xz - c2*s3*s_xx + s2*s_xy) + s2*(c2*c3*s_yz - c2*s3*s_xy + s2*s_yy)

        # XY
        arr[i*nitem + 3] = c1*c2*(-c2*s1*s_yy + s_xy*(c1*c3 - s1*s2*s3) + s_yz*(c1*s3 + c3*s1*s2)) + (-c1*c3*s2 + s1*s3)*(-c2*s1*s_yz + s_xz*(c1*c3 - s1*s2*s3) + s_zz*(c1*s3 + c3*s1*s2)) + (c1*s2*s3 + c3*s1)*(-c2*s1*s_xy + s_xx*(c1*c3 - s1*s2*s3) + s_xz*(c1*s3 + c3*s1*s2))

        # YZ
        arr[i*nitem + 4] = c2*c3*(c1*c2*s_yz + s_xz*(c1*s2*s3 + c3*s1) + s_zz*(-c1*c3*s2 + s1*s3)) - c2*s3*(c1*c2*s_xy + s_xx*(c1*s2*s3 + c3*s1) + s_xz*(-c1*c3*s2 + s1*s3)) + s2*(c1*c2*s_yy + s_xy*(c1*s2*s3 + c3*s1) + s_yz*(-c1*c3*s2 + s1*s3))

        # XZ
        arr[i*nitem + 5] = c2*c3*(-c2*s1*s_yz + s_xz*(c1*c3 - s1*s2*s3) + s_zz*(c1*s3 + c3*s1*s2)) - c2*s3*(-c2*s1*s_xy + s_xx*(c1*c3 - s1*s2*s3) + s_xz*(c1*s3 + c3*s1*s2)) + s2*(-c2*s1*s_yy + s_xy*(c1*c3 - s1*s2*s3) + s_yz*(c1*s3 + c3*s1*s2))


def read_nodal_values(filename, uint8 [::1] celltypes,
                      int64_t [::1] ele_ind_table,
                      int64_t [::1] offsets,
                      int64_t [::1] cells,
                      int nitems,
                      int npoints,
                      int [::1] nodstr,
                      int [::1] etype,
                      int [::1] element_type,
                      int result_index,
                      int64_t ptr_off):
    """Read nodal results from ANSYS directly into a numpy array

    element_type : int [::1] np.ndarray
        Array of ANSYS element types.

    ptr_off : int64_t
        Pointer offset

    result_index : int
        EMS - 0 : misc. data
        ENF - 1 : nodal forces
        ENS - 2 : nodal stresses
        ENG - 3 : volume and energies
        EGR - 4 : nodal gradients
        EEL - 5 : elastic strains
        EPL - 6 : plastic strains
        ECR - 7 : creep strains
        ETH - 8 : thermal strains
        EUL - 9 : euler angles
        EFX - 10 : nodal fluxes
        ELF - 11 : local forces
        EMN - 12 : misc. non-sum values
        ECD - 13 : element current densities
        ENL - 14 : nodal nonlinear data
        EHC - 15 : calculated heat
        EPT - 16 : element temperatures
        ESF - 17 : element surface stresses
        EDI - 18 : diffusion strains
        ETB - 19 : ETABLE items
        ECT - 20 : contact data
        EXY - 21 : integration point locations
        EBA - 22 : back stresses
        ESV - 23 : state variables
        MNL - 24 : material nonlinear record
    """
    cdef int64_t i, j, k, ind, nread, offset
    cdef int64_t ncells = ele_ind_table.size

    cdef bytes py_bytes = filename.encode()
    cdef char* c_filename = py_bytes
    cdef ifstream* binfile = new ifstream(c_filename, binary)

    cdef int [::1] ncount = np.zeros(npoints, ctypes.c_int32)

    # point data
    cdef double [:, ::1] data = np.zeros((npoints, nitems), np.float64)

    # temp buffer to hold data read from element
    cdef double [:, ::1] bufferdata = np.zeros((20, nitems), np.float64)

    cdef int64_t ele_table, nnode_elem
    cdef int ptr_result, skip
    cdef int c = 0
    cdef uint8 celltype
    for i in range(ncells):

        # read element data
        nnode_elem = nodstr[etype[i]]
        if ele_ind_table[i] == 0:  # element contains no data
            continue
        else:
            skip = read_element_result(binfile, ele_ind_table[i] + ptr_off,
                                       result_index, nnode_elem, nitems,
                                       &bufferdata[0, 0])
            if skip:
                continue

        # Get the nodes in the element
        celltype = celltypes[i]
        offset = offsets[i] + 1

        if celltype == VTK_LINE:
            read_element(cells, offset, ncount, data, bufferdata, nitems, 2)
        elif celltype == VTK_TRIANGLE:  # untested
            read_element(cells, offset, ncount, data, bufferdata, nitems, 3)
        elif celltype == VTK_QUAD or celltype == VTK_QUADRATIC_QUAD:
            read_element(cells, offset, ncount, data, bufferdata, nitems, 4)
        elif celltype == VTK_HEXAHEDRON:
            read_element(cells, offset, ncount, data, bufferdata, nitems, 8)
        elif celltype == VTK_PYRAMID:
            read_element(cells, offset, ncount, data, bufferdata, nitems, 5)
        elif celltype == VTK_TETRA:  # dependent on element type
            if nodstr[etype[i]] == 4:
                read_element(cells, offset, ncount, data, bufferdata, nitems, 4)
            else:
                read_tetrahedral(cells, offset, ncount, data, bufferdata, nitems)
        elif celltype == VTK_WEDGE:
            read_wedge(cells, offset, ncount, data, bufferdata, nitems)
    del binfile

    return np.asarray(data), np.asarray(ncount)


def read_nodal_values_dist(filename,
                           uint8 [::1] celltypes,
                           int64_t [::1] ele_ind_table,
                           int64_t [::1] offsets,
                           int64_t [::1] cells,
                           int nitems,
                           int npoints,
                           int [::1] nodstr,
                           int [::1] etype,
                           int [::1] element_type,
                           int result_index,
                           int64_t ptr_off,
                           int [::1] ncount,
                           double [:, ::1] data,
                           int c,
):
    """Read nodal results from a distributed result directly into a numpy array.

    element_type : int [::1] np.ndarray
        Array of ANSYS element types.

    ptr_off : int64_t
        Pointer offset

    result_index : int
        EMS - 0 : misc. data
        ENF - 1 : nodal forces
        ENS - 2 : nodal stresses
        ENG - 3 : volume and energies
        EGR - 4 : nodal gradients
        EEL - 5 : elastic strains
        EPL - 6 : plastic strains
        ECR - 7 : creep strains
        ETH - 8 : thermal strains
        EUL - 9 : euler angles
        EFX - 10 : nodal fluxes
        ELF - 11 : local forces
        EMN - 12 : misc. non-sum values
        ECD - 13 : element current densities
        ENL - 14 : nodal nonlinear data
        EHC - 15 : calculated heat
        EPT - 16 : element temperatures
        ESF - 17 : element surface stresses
        EDI - 18 : diffusion strains
        ETB - 19 : ETABLE items
        ECT - 20 : contact data
        EXY - 21 : integration point locations
        EBA - 22 : back stresses
        ESV - 23 : state variables
        MNL - 24 : material nonlinear record

    data : double [:, ::1] np.ndarray
        Data for each individual node.

    c : int
        Global index counter.  Keeps track of the cell location
        between multiple result files.

    """
    # number of cells in this distributed result file
    cdef int64_t ncells = ele_ind_table.size

    # open this result file
    cdef bytes py_bytes = filename.encode()
    cdef char* c_filename = py_bytes
    cdef ifstream* binfile = new ifstream(c_filename, binary)

    # temp buffer to hold data read from element
    cdef int64_t i, j, k, ind, nread, offset
    cdef double [:, ::1] bufferdata = np.zeros((20, nitems), np.float64)
    cdef int64_t ele_table, nnode_elem
    cdef int ptr_result, skip
    cdef uint8 celltype
    for i in range(ncells):

        # read element data
        nnode_elem = nodstr[etype[i]]  # global
        if ele_ind_table[i] == 0:  # element contains no data
            c += 1  # global solution cell index
            continue
        else:
            skip = read_element_result(binfile, ele_ind_table[i] + ptr_off,
                                       result_index, nnode_elem, nitems,
                                       &bufferdata[0, 0])
            if skip:
                c += 1  # global solution cell index
                continue

        # Get number of nodes in the global element
        celltype = celltypes[c]
        offset = offsets[c] + 1
        # NOTE: value at offsets[c] is the number of points in cell

        if celltype == VTK_LINE:
            read_element(cells, offset, ncount, data, bufferdata, nitems, 2)
        elif celltype == VTK_TRIANGLE:  # untested
            read_element(cells, offset, ncount, data, bufferdata, nitems, 3)
        elif celltype == VTK_QUAD or celltype == VTK_QUADRATIC_QUAD:
            read_element(cells, offset, ncount, data, bufferdata, nitems, 4)
        elif celltype == VTK_HEXAHEDRON:
            read_element(cells, offset, ncount, data, bufferdata, nitems, 8)
        elif celltype == VTK_PYRAMID:
            read_element(cells, offset, ncount, data, bufferdata, nitems, 5)
        elif celltype == VTK_TETRA:  # dependent on element type
            if nodstr[etype[i]] == 4:
                read_element(cells, offset, ncount, data, bufferdata, nitems, 4)
            else:
                read_tetrahedral(cells, offset, ncount, data, bufferdata, nitems)
        elif celltype == VTK_WEDGE:
            read_wedge(cells, offset, ncount, data, bufferdata, nitems)

        c += 1  # global solution cell index

    # must delete to manually collect
    del binfile

    # have to increment again
    # return c + 1
    return c

# indices of a wedge must be reordered (see _parser.store_weg)
cdef int64_t [6] wedge_ind
wedge_ind[0] = 2
wedge_ind[1] = 1
wedge_ind[2] = 0
wedge_ind[3] = 6
wedge_ind[4] = 5
wedge_ind[5] = 4

cdef inline void read_wedge(int64_t [::1] cells, int64_t index, int [::1] ncount,
                           float_or_double [:, ::1] data,
                            float_or_double [:, ::1] bufferdata,
                           int nitems) nogil:
    """
    [0, 1, 2, 2, 3, 4, 5, 5]
    [0, 1, 2,  , 4, 5, 6,  ]
    """
    cdef int64_t i, j, cell, idx
    cdef int nread = nitems*8
    
    for i in range(6):
        cell = cells[index + i]
        ncount[cell] += 1
        idx = wedge_ind[i]
        for j in range(nitems):
            data[cell, j] += bufferdata[idx, j]


# indices of a 186 tetrahedral must be reordered (see _parser.StoreWeg)
cdef int64_t [6] tet_ind
tet_ind[0] = 0
tet_ind[1] = 1
tet_ind[2] = 2
tet_ind[3] = 4

cdef inline void read_tetrahedral(int64_t [::1] cells, int64_t index, int [::1] ncount,
                                  float_or_double [:, ::1] data, float_or_double [:, ::1] bufferdata,
                                  int nitems) nogil:
    """
    # see documentation at _parser.StoreTet
    """
    cdef int64_t i, j, cell, idx
    cdef int nread

    nread = nitems*5

    for i in range(4):
        cell = cells[index + i]
        ncount[cell] += 1
        idx = tet_ind[i]
        for j in range(nitems):
            data[cell, j] += bufferdata[idx, j]


cdef inline void read_element(int64_t [::1] cells, int64_t index, int [::1] ncount,
                              float_or_double [:, ::1] data,
                              float_or_double [:, ::1] bufferdata,
                              int nitems, int nnode) nogil:
    """
    Reads a generic element type in a linear fashion.  Works for:
    Hexahedron 95 or 186
    Pyramid 95 or 186
    Tetrahedral 187
    """
    cdef int64_t i, j, cell, idx

    for i in range(nnode):
        cell = cells[index + i]
        ncount[cell] += 1
        for j in range(nitems):
            data[cell, j] += bufferdata[i, j]


def read_array(filename, int ptr, int nterm, int neqn, int [::1] const):
    """Reads stiffness or mass matrices from ANSYS fortran files

    Parameters
    ----------
    filename : string
        Full filename
        
    ptr: int
        Pointer to start of block
        
    nterm : int
        Number of terms to read.
        
    neqn : int
        Number of equations
        
    const : numpy int array
        If DOF is fixed

    Returns
    -------
    rows : numpy int32 array
        Row indices
    
    cols : numpy int32 array
        Column indices
    
    data : numpy double array
        Data belonging to (row, col)
    
    diag : numpy int32 array
        Indices along the diag (diag[i], diag[i])
    
    data_diag : numpy double array
        Data belonging to the diag entries
    """

    cdef int i, j, k, ind
    cdef bytes buf
    with open(filename, "rb") as f:
        f.seek(ptr*4)
        buf = f.read((neqn*6 + nterm*3)*4)

    cdef char * p = buf # python to c character array pointer
    cdef int loc = 0 # location long buffer

    # Number of terms is the number of terms stored in the upper triangle
    cdef int [::1] krow = np.empty(nterm, np.int32)
    cdef int [::1] kcol = np.empty(nterm, np.int32)
    cdef double [::1] kdata = np.empty(nterm)

    cdef int [::1] kdiag = np.empty(neqn, np.int32)
    cdef double [::1] kdata_diag = np.empty(neqn)

    cdef int c = 0 # index counter
    cdef int d = 0 # data counter
    cdef int c_diag = 0 # diag counter
    cdef int row, col, nitems, intval
    cdef double val

    for i in range(neqn):
        col = i

        # number of items to read
        nitems = get_int(&p[loc]); loc += 4
        loc += 4

        # Read indices
        for j in range(nitems):
            # get row number
            row = get_int(&p[loc]) - 1; loc += 4 # convert to c indexing

            if row < col:
                krow[c] = row
                kcol[c] = col
            else:
                krow[c] = col
                kcol[c] = row
            c += 1

        loc += 12
            
        # Read data
        for j in range(nitems):
            # Store data
            kdata[d] = get_double(&p[loc]); loc += 8
            d += 1

        # seek past end of data
        loc += 4

    return np.asarray(krow)[:c], np.asarray(kcol)[:c], np.asarray(kdata)[:c]


def sort_nodal_eqlv(int neqn, int [::1] neqv, int [::1] ndof):
    """Read in full file details required for the assembly of the mass
    and stiffness matrices.

    The reference arrays are sorted by default, though this increases
    the bandwidth of the mass and stiffness matrices.

    Parameters
    ----------
    neqn : int
        Number of equations in full file.

    ndof : int [::1]
        Degrees of freedom for each node.

    neqv : int [::1]
        Nodal equivalance array.

    Returns
    -------
    dof_ref: numpy np.int32 array
        Sorted degree of freedom reference array.
        
    index_arr : numpy np.int32 array
        Index array to sort rows and columns.

    """
    cdef int i, j, ind

    # create sorting array
    cdef int nnodes = ndof.size
    cdef int [::1] cumdof = np.empty(nnodes, np.int32)
    cdef int csum = 0
    for i in range(nnodes):
        cumdof[i] = csum
        csum += ndof[i]
        
    cdef int [::1] s_neqv_dof = np.empty(neqn, np.int32)
    cdef int [::1] nref = np.empty(neqn, np.int32)
    cdef int [::1] dref = np.empty(neqn, np.int32)
    cdef int c = 0
    cdef int val
    for i in range(nnodes):
        val = neqv[i]
        for j in range(ndof[i]):
            nref[c] = val
            dref[c] = j
            c += 1

    # sort nodal equivalance array
    cdef int [::1] sidx = np.argsort(neqv).astype(np.int32)
    cdef int [::1] ndof_sort = np.empty(nnodes, np.int32)
    for i in range(nnodes):
        ndof_sort[i] = ndof[sidx[i]]

    cdef int d = 0
    # create an index array.  This tells the array readers where
    # to place each row and col when it's sorted
    cdef int [::1] index_arr = np.empty(neqn, np.int32)
    for i in range(nnodes):
        ind = sidx[i]
        c = cumdof[ind]
        for j in range(ndof[ind]):
            s_neqv_dof[d] = c + j
            index_arr[c + j] = d
            d += 1

    # sort node and dof references
    cdef int [:, ::1] dof_ref = np.empty((neqn, 2), np.int32)
    for i in range(neqn):
        ind = s_neqv_dof[i]
        dof_ref[i, 0] = nref[ind]
        dof_ref[i, 1] = dref[ind]

    return np.asarray(dof_ref), np.asarray(index_arr), np.asarray(nref), \
           np.asarray(dref)


def tensor_arbitrary(double [:, ::1] stress, double [:, :] trans):
    """Rotates a 3D stress tensor by theta about the Z axis

    Notes:
    -----
    Used
    from sympy import Matrix, symbols
    s_xx, s_yy, s_zz, s_xy, s_yz, s_xz = symbols('s_xx s_yy s_zz s_xy s_yz s_xz')
    c0, c1, c2, c3, c4, c5, c6, c7, c8 = symbols('c0 c1 c2 c3 c4 c5 c6 c7 c8')
    
    R = Matrix([[c0, c1, c2], [c3, c4, c5], [c6, c7, c8]])
    tensor = Matrix([[s_xx, s_xy, s_xz], [s_xy, s_yy, s_yz], [s_xz, s_yz, s_zz]])
    R*tensor*R.T
    """
    cdef int nnode = stress.shape[0]
    cdef int i

    cdef uint8 [::1] isnan = np.zeros(nnode, np.uint8)
    cdef double s_xx, s_yy, s_zz, s_xy, s_yz, s_xz
    cdef double c0 = trans[0, 0]
    cdef double c1 = trans[0, 1]
    cdef double c2 = trans[0, 2]
    cdef double c3 = trans[1, 0]
    cdef double c4 = trans[1, 1]
    cdef double c5 = trans[1, 2]
    cdef double c6 = trans[2, 0]
    cdef double c7 = trans[2, 1]
    cdef double c8 = trans[2, 2]

    cdef double r0, r1, r2, r3, r4, r5, r6, r7, r8

    for i in range(nnode):
        s_xx = stress[i, 0]
        if npy_isnan(s_xx):  # skip
            isnan[i] = 1
        else:
            s_yy = stress[i, 1]
            s_zz = stress[i, 2]
            s_xy = stress[i, 3]
            s_yz = stress[i, 4]
            s_xz = stress[i, 5]

        r0 = c0*(c0*s_xx + c1*s_xy + c2*s_xz) + c1*(c0*s_xy + c1*s_yy + c2*s_yz) + c2*(c0*s_xz + c1*s_yz + c2*s_zz)

        r1 = c3*(c0*s_xx + c1*s_xy + c2*s_xz) + c4*(c0*s_xy + c1*s_yy + c2*s_yz) + c5*(c0*s_xz + c1*s_yz + c2*s_zz)

        r4 = c3*(c3*s_xx + c4*s_xy + c5*s_xz) + c4*(c3*s_xy + c4*s_yy + c5*s_yz) + c5*(c3*s_xz + c4*s_yz + c5*s_zz)

        r5 = c6*(c3*s_xx + c4*s_xy + c5*s_xz) + c7*(c3*s_xy + c4*s_yy + c5*s_yz) + c8*(c3*s_xz + c4*s_yz + c5*s_zz)

        r6 = c0*(c6*s_xx + c7*s_xy + c8*s_xz) + c1*(c6*s_xy + c7*s_yy + c8*s_yz) + c2*(c6*s_xz + c7*s_yz + c8*s_zz)

        r8 = c6*(c6*s_xx + c7*s_xy + c8*s_xz) + c7*(c6*s_xy + c7*s_yy + c8*s_yz) + c8*(c6*s_xz + c7*s_yz + c8*s_zz)

        stress[i, 0] = r0
        stress[i, 1] = r4
        stress[i, 2] = r8
        stress[i, 3] = r1
        stress[i, 4] = r5
        stress[i, 5] = r6

    return np.asarray(isnan, np.bool)


def tensor_strain_arbitrary(double [:, ::1] stress, double [:, :] trans):
    """Rotates a 3D strain tensor using an arbitrary transformation matrix

    Notes:
    -----
    Used
    from sympy import Matrix, symbols
    s_xx, s_yy, s_zz, s_xy, s_yz, s_xz = symbols('s_xx s_yy s_zz s_xy s_yz s_xz')
    c0, c1, c2, c3, c4, c5, c6, c7, c8 = symbols('c0 c1 c2 c3 c4 c5 c6 c7 c8')
    
    R = Matrix([[c0, c1, c2], [c3, c4, c5], [c6, c7, c8]])
    tensor = Matrix([[s_xx, s_xy, s_xz], [s_xy, s_yy, s_yz], [s_xz, s_yz, s_zz]])
    R*tensor*R.T

    Additional note:
    Shear components must be pre multiplied by 0.5 and then
    post-multiplied by 2 due to the classical definition of shear
    strain being half of tonsorial shear strain.

    See:
    http://web.mit.edu/course/3/3.11/www/modules/trans.pdf

    """
    cdef int nnode = stress.shape[0]
    cdef int i

    cdef uint8 [::1] isnan = np.zeros(nnode, np.uint8)
    cdef double s_xx, s_yy, s_zz, s_xy, s_yz, s_xz
    cdef double c0 = trans[0, 0]
    cdef double c1 = trans[0, 1]
    cdef double c2 = trans[0, 2]
    cdef double c3 = trans[1, 0]
    cdef double c4 = trans[1, 1]
    cdef double c5 = trans[1, 2]
    cdef double c6 = trans[2, 0]
    cdef double c7 = trans[2, 1]
    cdef double c8 = trans[2, 2]

    cdef double r0, r1, r2, r3, r4, r5, r6, r7, r8

    for i in range(nnode):
        s_xx = stress[i, 0]
        if npy_isnan(s_xx):  # skip
            isnan[i] = 1
        else:
            s_yy = stress[i, 1]
            s_zz = stress[i, 2]
            s_xy = stress[i, 3]*0.5
            s_yz = stress[i, 4]*0.5
            s_xz = stress[i, 5]*0.5

        r0 = c0*(c0*s_xx + c1*s_xy + c2*s_xz) + c1*(c0*s_xy + c1*s_yy + c2*s_yz) + c2*(c0*s_xz + c1*s_yz + c2*s_zz)

        r1 = c3*(c0*s_xx + c1*s_xy + c2*s_xz) + c4*(c0*s_xy + c1*s_yy + c2*s_yz) + c5*(c0*s_xz + c1*s_yz + c2*s_zz)

        r4 = c3*(c3*s_xx + c4*s_xy + c5*s_xz) + c4*(c3*s_xy + c4*s_yy + c5*s_yz) + c5*(c3*s_xz + c4*s_yz + c5*s_zz)

        r5 = c6*(c3*s_xx + c4*s_xy + c5*s_xz) + c7*(c3*s_xy + c4*s_yy + c5*s_yz) + c8*(c3*s_xz + c4*s_yz + c5*s_zz)

        r6 = c0*(c6*s_xx + c7*s_xy + c8*s_xz) + c1*(c6*s_xy + c7*s_yy + c8*s_yz) + c2*(c6*s_xz + c7*s_yz + c8*s_zz)

        r8 = c6*(c6*s_xx + c7*s_xy + c8*s_xz) + c7*(c6*s_xy + c7*s_yy + c8*s_yz) + c8*(c6*s_xz + c7*s_yz + c8*s_zz)

        stress[i, 0] = r0
        stress[i, 1] = r4
        stress[i, 2] = r8
        stress[i, 3] = r1*2
        stress[i, 4] = r5*2
        stress[i, 5] = r6*2

    return np.asarray(isnan, np.bool)


def tensor_rotate_z(double [:, :] stress, float theta_z):
    """Rotates a 3D stress tensor by theta about the Z axis

    Notes:
    -----
    Used 
    from sympy import Matrix, symbols
    c, s, s_xx, s_yy, s_zz, s_xy, s_yz, s_xz = symbols('c s s_xx s_yy s_zz s_xy s_yz s_xz')

    R = Matrix([[c, -s, 0], [s, c, 0], [0, 0, 1]])
    tensor = Matrix([[s_xx, s_xy, s_xz], [s_xy, s_yy, s_yz], [s_xz, s_yz, s_zz]])
    R*tensor*R.T
    """
    cdef int nnode = stress.shape[0]
    cdef int i
    cdef uint8 [::1] isnan = np.zeros(nnode, np.uint8)
    cdef double c, s, s_xx, s_yy, s_zz, s_xy, s_yz, s_xz

    c = cos(theta_z)
    s = sin(theta_z)

    for i in range(nnode):
        s_xx = stress[i, 0]
        if npy_isnan(s_xx):
            isnan[i] = 1
        else:
            s_yy = stress[i, 1]
            s_zz = stress[i, 2]
            s_xy = stress[i, 3]
            s_yz = stress[i, 4]
            s_xz = stress[i, 5]

        stress[i, 0] = c*(c*s_xx - s*s_xy) - s*(c*s_xy - s*s_yy)
        stress[i, 1] = c*(c*s_yy + s*s_xy) + s*(c*s_xy + s*s_xx)
        stress[i, 2] = s_zz
        stress[i, 3] = c*(c*s_xy - s*s_yy) + s*(c*s_xx - s*s_xy)
        stress[i, 4] = c*s_yz + s*s_xz
        stress[i, 5] = c*s_xz - s*s_yz

    return np.asarray(isnan, dtype=np.bool)


def compute_principal_stress(double [:, ::1] stress):
    """Returns the principal stresses based on component stresses.

    Parameters
    ----------
    stress : numpy.ndarray (double)
        Stresses at Sx Sy Sz Sxy Syz Sxz averaged at each corner node.

    Returns
    -------
    pstress : numpy.ndarray
        Principal stresses, stress intensity, and equivalant stress.
        [sigma1, sigma2, sigma3, sint, seqv]

    Notes
    -----
    ANSYS equivalant of:
    PRNSOL, S, PRIN

    Which returns:
    S1, S2, S3 principal stresses, SINT stress intensity, and SEQV
    equivalent stress.
    """
    # reshape the stress array into 3x3 stress tensor arrays
    cdef int nnode = stress.shape[0]
    cdef double [:, :, ::1] stress_tensor = np.empty((nnode, 3, 3), np.float64)
    cdef double s_xx, x_yy, s_zz, s_xy, s_yz, s_xz
    cdef int i

    cdef uint8 [::1] isnan = np.zeros(nnode, np.uint8)

    for i in range(nnode):
        s_xx = stress[i, 0]
        if npy_isnan(s_xx):
            s_xx = 0
            s_yy = 0
            s_zz = 0
            s_xy = 0
            s_yz = 0
            s_xz = 0
            isnan[i] = 1
        else:
            s_yy = stress[i, 1]
            s_zz = stress[i, 2]
            s_xy = stress[i, 3]
            s_yz = stress[i, 4]
            s_xz = stress[i, 5]

        # populate lower triangle of stress tensor
        stress_tensor[i, 0, 0] = s_xx
        # stress_tensor[i, 0, 1] = s_xy
        # stress_tensor[i, 0, 2] = s_xz
        stress_tensor[i, 1, 0] = s_xy
        stress_tensor[i, 1, 1] = s_yy
        # stress_tensor[i, 1, 2] = s_yz
        stress_tensor[i, 2, 0] = s_xz
        stress_tensor[i, 2, 1] = s_yz
        stress_tensor[i, 2, 2] = s_zz

    # compute principle stresses
    # w =  np.linalg.eigvalsh(stress_tensor)  # default uses lower triangle
    # wrapped lapack libaray (slightly faster than above)$
    w = np.linalg._umath_linalg.eigvalsh_lo(stress_tensor)
    w[:, ::-1].sort(1)

    temp = np.empty((nnode, 5), np.float64)
    temp[:, :3] = w

    cdef double [:, ::1] pstress = temp
    cdef double p1, p2, p3, c1, c2, c3

    # compute stress intensity and von mises (equivalent) stress
    for i in range(nnode):
        p1 = pstress[i, 0]
        p2 = pstress[i, 1]
        p3 = pstress[i, 2]

        c1 = fabs(p1 - p2)
        c2 = fabs(p2 - p3)
        c3 = fabs(p3 - p1)

        if c1 > c2:
            if c1 > c3:
                pstress[i, 3] = c1
            else:
                pstress[i, 3] = c3
        else:
            if c2 > c3:
                pstress[i, 3] = c2
            else:
                pstress[i, 3] = c3

        pstress[i, 4] = sqrt(0.5*(c1**2 + c2**2 + c3**2))

    return np.asarray(pstress), np.asarray(isnan, np.bool)


def affline_transform(float_or_double [:, ::1] points, float_or_double [:, ::1] t):
    """ Rigidly transforms points based on a 4x4 transform matrix """
    cdef int npoints = points.shape[0]
    cdef int i, j
    cdef float_or_double x, y, z
    cdef float_or_double xnew, ynew, znew

    cdef float_or_double t00 = t[0, 0]
    cdef float_or_double t01 = t[0, 1]
    cdef float_or_double t02 = t[0, 2]
    cdef float_or_double t03 = t[0, 3]

    cdef float_or_double t10 = t[1, 0]
    cdef float_or_double t11 = t[1, 1]
    cdef float_or_double t12 = t[1, 2]
    cdef float_or_double t13 = t[1, 3]

    cdef float_or_double t20 = t[2, 0]
    cdef float_or_double t21 = t[2, 1]
    cdef float_or_double t22 = t[2, 2]
    cdef float_or_double t23 = t[2, 3]

    for i in prange(npoints, nogil=True):
        x = points[i, 0]
        y = points[i, 1]
        z = points[i, 2]

        points[i, 0] = t00*x + t01*y + t02*z + t03
        points[i, 1] = t10*x + t11*y + t12*z + t13
        points[i, 2] = t20*x + t21*y + t22*z + t23


cdef inline int cell_lookup(uint8 celltype) nogil:
    if celltype == VTK_HEXAHEDRON or celltype == VTK_QUADRATIC_HEXAHEDRON:
        return 8
    elif celltype == VTK_TETRA or celltype == VTK_QUADRATIC_TETRA:
        return 4
    elif celltype == VTK_PYRAMID or celltype == VTK_QUADRATIC_PYRAMID:
        return 5
    elif celltype == VTK_WEDGE or celltype == VTK_QUADRATIC_WEDGE:
        return 6


def cells_with_all_nodes(index_type [::1] offset, index_type [::1] cells,
                         uint8 [::1] celltypes, uint8 [::1] point_mask):
    """
    Updates mask of cells containing all points in the point indices
    or mask.
    """
    cdef int ncells = celltypes.size
    cdef uint8 celltype
    cdef int ncell_points, i, j
    cdef index_type cell_offset
    cdef uint8 [::1] cell_mask = np.ones(ncells, np.uint8)

    with nogil:
        for i in range(ncells):
            celltype = celltypes[i]
            ncell_points = cell_lookup(celltype)
            cell_offset = offset[i] + 1
            for j in range(cell_offset, cell_offset + ncell_points):
                if point_mask[cells[j]] != 1:
                    cell_mask[i] = 0

    return np.asarray(cell_mask, dtype=np.bool)


def cells_with_any_nodes(index_type [::1] offset, index_type [::1] cells,
                         uint8 [::1] celltypes, uint8 [::1] point_mask):
    """
    Updates mask of cells containing at least one point in the point
    indices or mask.
    """
    cdef int ncells = celltypes.size
    cdef uint8 celltype
    cdef int ncell_points
    cdef index_type cell_offset
    cdef int i, j

    cdef uint8 [::1] cell_mask = np.zeros(ncells, np.uint8)

    with nogil:
        for i in range(ncells):
            celltype = celltypes[i]
            ncell_points = cell_lookup(celltype)
            cell_offset = offset[i] + 1
            for j in range(cell_offset, cell_offset + ncell_points):
                if point_mask[cells[j]] == 1:
                    cell_mask[i] = 1
                    break

    return np.asarray(cell_mask, dtype=np.bool)


def midside_mask(uint8 [::1] celltypes, index_type [::1] cells,
                 index_type [::1] offset, int npoints):
    """Returns a mask of midside nodes

    Parameters
    ----------
    celltypes : uint8 [::1]
        VTK style celltype array

    cells : int32 or int64 [::1]
        VTK style cell array

    offset : int32 or int64 [::1]
        VTK style offset array

    npoints : int
        Number of points

    Returns
    -------
    mask : bool np.ndarray
        True when a midside node.
    """
    cdef uint8 [::1] mask = np.zeros(npoints, ctypes.c_uint8)
    cdef int i, j, c
    cdef int ncells = celltypes.size
    cdef uint8 celltype

    for i in range(ncells):
        # get start location of each cell
        c = offset[i] + 1
        celltype = celltypes[i]
    
        if celltype == VTK_QUADRATIC_TETRA:
            for j in range(c + 4, c + 10):
                mask[cells[j]] = 1

        elif celltype == VTK_QUADRATIC_PYRAMID:
            for j in range(c + 5, c + 13):
                mask[cells[j]] = 1

        elif celltype == VTK_QUADRATIC_WEDGE:
            for j in range(c + 6, c + 15):
                mask[cells[j]] = 1

        elif celltype == VTK_QUADRATIC_HEXAHEDRON:   
            for j in range(c + 8, c + 20):
                mask[cells[j]] = 1

    # return as a bool array without copying
    return np.asarray(mask).view(np.bool)


def euler_cart_to_cyl(double [:, ::1] stress, double [::1] angles):
    """Convert stress tensors from cartesian to cyclindrical.

    Equations from
    https://github.com/jrwrigh/cart2cyl_notes/blob/master/main.pdf

    Parameters
    ----------
    stress : np.ndarray (np.double)
        ``n x 6`` stress tensor array of values
        s_xx, s_xy, s_yy, s_xz, s_yz, s_zz

    angles : np.ndarray
        Rotation angles.  Effectively arctan(y/x)

    Notes
    -----
    # used sympy to generate these equations
    >>> from sympy import Matrix, symbols
    >>> c, s = symbols('c s')
    >>> s_xx, s_xy, s_yy, s_xz, s_yz, s_zz = symbols('s_xx s_xy s_yy s_xz s_yz s_zz')
    >>> tensor = np.matrix([[s_xx, s_xy, s_xz],
                            [s_xy, s_yy, s_yz],
                            [s_xz, s_yz, s_zz]])
    >>> r1 = Matrix([[c, s, 0],
                     [-s, c, 0],
                     [0, 0, 1]])
    >>> r2 = Matrix([[c, -s, 0],
                     [s, c, 0],
                     [0, 0, 1]])
    >>> r1*tensor*r2
    """
    cdef double s_xx, s_xy, s_yy, s_xz, s_yz, s_zz  # stress tensors
    cdef double c_th, s_th
    cdef int i
    cdef int n_node = stress.shape[0]

    # check to prevent a segfault
    if angles.size != n_node:
        raise ValueError('Number of angles must match the number of stress values')

    for i in range(n_node):
        c_th = cos(angles[i])
        s_th = sin(angles[i])

        # grab the node component stresses
        s_xx = stress[i, 0]
        s_yy = stress[i, 1]
        s_zz = stress[i, 2]
        s_xy = stress[i, 3]
        s_yz = stress[i, 4]
        s_xz = stress[i, 5]

        # store rotated component stresses 
        # RR (was XX)
        stress[i, 0] = c_th**2*s_xx + 2*c_th*s_th*s_xy + s_th**2*s_yy

        # THETATHETA (was YY)
        stress[i, 1] = s_th**2*s_xx - 2*c_th*s_th*s_xy + c_th**2*s_yy
        

        # ZZ (same)
        # stress[i, 2] =
        # 

        # RTHETA (was XY)
        stress[i, 3] = c_th*s_th*(s_yy - s_xx) + (c_th**2 - s_th**2)*s_xy
        

        # THETAZ (was YZ)
        stress[i, 4] = -s_th*s_xz + c_th*s_yz

        # RZ (was XZ)
        stress[i, 5] = c_th*s_xz + s_th*s_yz


def break_apart_surface(double [:, ::1] points, int64_t [::1] faces, int n_faces,
                        int force_linear=True):
    """Break apart the faces of a vtk PolyData such that the points
    for each face are unique and each point is used only by one face.
    This leads to duplicate points, but allows multiple scalars per
    face.

    Parameters
    ----------
    points : double [:, ::1] np.ndarray
        Points from a pyvista.PolyData

    faces : int64 [::1] np.ndarray
        Faces from a ``pyvista.PolyData``.  Maximum of 8 points per face.

    n_faces : int
        Number of faces

    force_linear : bool, optional
        When ``True``, converts quadratic faces to their linear counterparts.

    Returns
    -------
    new_points : double [:, ::1] np.ndarray
        New points applicable for a pyvista.PolyData

    new_faces : int64 [::1] np.ndarray
        New faces from a pyvista.PolyData.  Same size as faces.

    orig_idx : int32 [::1] np.ndarray
        Relates indices of the new points to the original surface.

    """
    cdef double [:, ::1] new_points = np.empty((n_faces*8, 3))
    cdef int64_t [::1] new_faces = np.empty(faces.size, np.int64)
    cdef int32_t [::1] orig_idx = np.empty(n_faces*8, np.int32)

    cdef int face_arr_sz = faces.size
    cdef int i = 0
    cdef int c = 0
    cdef int cj = 0
    cdef int j, face_sz, c_add
    cdef int face = 0
    while i < face_arr_sz:
        face += 1
        face_sz = faces[i]

        c_add = 0  # additional counter if face_sz is less than actual face size
        if face_sz > 8:
            raise Exception
        if face_sz > 4 and force_linear:
            if face_sz == 6:
                face_sz = 3
                c_add = 3
            elif face_sz == 8:
                face_sz = 4
                c_add = 4

        new_faces[cj] = face_sz
        i += 1  # counter for faces
        cj += 1  # counter for new_faces
        for j in range(face_sz):
            f_idx = faces[i]
            new_faces[cj] = c
            new_points[c] = points[f_idx]
            orig_idx[c] = f_idx
            c += 1
            cj += 1
            i += 1
        i += c_add  # extra increment because we're not recording all faces

    return np.array(new_points[:c]), np.array(new_faces[:cj]), np.array(orig_idx[:c])




"""pyansys Element matricies file reader


This header is straight out of fdemat.inc and can be found online at
https://www.sharcnet.ca/Software/Ansys/16.2.3/en-us/help/ans_prog/Hlp_P_INT1_6.html

*comdeck,fdemat

c *** Copyright ANSYS.  All Rights Reserved.
c *** ansys, inc.

c     **********  description of element matrix file  **********
c
c *** mpg fdemat.inc < eoelem elostr eofini outelm elfini EmatAssemble sffini
c          eqprep sfform elstrt slvstr: emat file description
c
      character*8  EMATNM 
      parameter  (EMATNM='emat    ')

      LONGINT         ematfpL, ematfp
      integer         ematbk, ematut
      common /fdemat/ ematfpL, ematbk, ematut
      equivalence  (ematfp,ematfpL)

c   ********** common variable descriptions ***********
co  ematfpL      file position on file emat
co  ematbk       block number for file emat
co  ematut       file unit for file emat

c   See fddesc for documentation of how binary files are stored.
c  
c     **********  file format  **********

c        recid tells the identifier for this record.  Not all records will have
c             identifiers -- they are only indicated for those records whose 
c             record pointers are stored in the second file header.

c        type tells what kind of information is stored in this record:
c             i - integer
c             dp - double precision
c             cmp - complex

c        nrec tells how many records of this description are found here

c        lrec tells how long the records are (how many items are stored)

c recid    type    nrec    lrec     contents

c  ---      i       1      100      standard ANSYS file header (see binhed for 
c                                   details of header contents)

c  ---      i       1       40      .EMAT FILE HEADER
c                        
c                                    fun02,    nume,   numdof,    lenu,  lenbac,
c                                     maxn, nlgeEMA,  sstEMAT,  nodref,   lumpm,
c                                    kygst,    kygm,     kycd,   kygss,   kygaf,
c                                    kygrf,       0,Glblenbac, ptrGBkl, ptrGBkh,
c                                  ptrElmh, ptrFSTh,  ptrLSTh, ptrBITh, ptrEHDh,
c                                  ptrIDXh,   numCE,  maxLeng,  ptrCEl,  ptrCEh,
c                                   ptrDOF,  ptrBAC,  ptrELMl, ptrFSTl, ptrLSTl,
c                                  ptrBITl, ptrEHDl,  ptrIDXl, ptrendH, ptrendL

c                                each item in header is described below:

c                                   fun02  - unit number (emat file is 2)
c                                   nume   - number of elements
c                                   numdof - number of dofs per node
c                                   lenu   - total DOFs of model
c                                   lenbac - number of nodes
c                                   maxn   - maximum node number
c                                   nlgeEMA  = 0 - nlgeom is OFF the time this Emat file is created
c                                              1 - nlgeom is ON the time this Emat file is created
c                                   sstEMAT  = 0 - sstif key is OFF the time this Emat file is created
c                                              1 - sstif key is ON the time this Emat file is created
c                                                  this key is for internal use only
c                                   nodref - actual number of nodes referenced
c                                   lumpm  - lumped mass key 
c                                            = 0 - default matrix type
c                                            = 1 - lumped
c                                   kygst  - global stiffness matrix calculate 
c                                            key 
c                                            = 0 - do not calculate
c                                            = 1 - calculate
c                                   kygm   - global mass matrix calculate key 
c                                            = 0 - do not calculate
c                                            = 1 - calculate
c                                   kycd   - global damping matrix calculate key
c                                            = 0 - do not calculate
c                                            = 1 - calculate
c                                   kygss  - global stress stiffening matrix 
c                                            calculate key 
c                                            = 0 - do not calculate
c                                            = 1 - calculate
c                                   kygaf  - global applied force vector 
c                                            calculate key
c                                            = 0 - do not calculate
c                                            = 1 - calculate
c                                   kygrf  - global restoring force vector 
c                                            calculate key (Newton-Raphson only)
c                                            = 0 - do not calculate
c                                            = 1 - calculate
c                                   0      - position not used
c                                   Glblenbac - global global number of nodes (== lenbac unless using
c                                               Distributed Ansys)
c                                   ptrGBkl- low pointer to global nodal equivalence table
c                                   ptrGBkh- high pointer to global nodal equivalence table
c                                   ptrELMh- high pointer to element equivalence table
c                                   ptrFSTh- high pointer to first element at a
c                                            DOF table 
c                                   ptrLSTh- high pointer to last element at a
c                                            DOF table
c                                   ptrBITh- high pointer to dof bits
c                                   ptrEHDh- high pointer to the start of the
c                                            element matrices
c                                   ptrIDXh- high pointer to element matrices
c                                            index table
c                                   numCE  - number of internal CEs
c                                   maxLeng- maximum length of any internal CE
c                                   ptrCEl - low pointer to internal CE list
c                                   ptrCEh - high pointer to internal CE list
c                                   ptrDOF - pointer to degrees of freedom per 
c                                            node used in model
c                                   ptrBAC - pointer to nodal equivalence table

c                                   ptrELMl- Low pointer to element equivalence 
c                                            table
c                                   ptrFSTl- Low pointer to first element at a
c                                            DOF table 
c                                   ptrLSTl- Low pointer to last element at a
c                                            DOF table
c                                   ptrBITl- Low pointer to dof bits
c                                   ptrEHDl- Low pointer to the start of the
c                                            element matrices
c                                   ptrIDXl- Low pointer to element matrices
c                                            index table

c                                   ptrendH- High pointer to end of file
c                                   ptrendL- Low  pointer to end of file

c       Note: the analysis type sets the global calculate keys.

c  ---     dp       1       20      Time information
c                                   
c                                   timval, timinc, frqval, timbeg, timend,
c                                      0.0,    0.0,    0.0,    0.0,    0.0,
c                                      0.0,    0.0,    0.0,    0.0,    0.0,
c                                      0.0,    0.0,    0.0,    0.0,    0.0,

c                                  each item is described below:

c                                   timval - the current time
c                                   timinc - the time increment
c                                   frqval - the current frequency (from a 
c                                            harmonic analysis)
c                                   timbeg - the start time for the analysis
c                                   timend - the end time for the analysis
c                                   0.0    - position not used
c                                   0.0    - position not used
c                                   0.0    - position not used
c                                   0.0    - position not used
c                                   0.0    - position not used
c                                   0.0    - position not used
c                                   0.0    - position not used
c                                   0.0    - position not used
c                                   0.0    - position not used
c                                   0.0    - position not used
c                                   0.0    - position not used
c                                   0.0    - position not used
c                                   0.0    - position not used
c                                   0.0    - position not used
c                                   0.0    - position not used

c  DOF      i       1    numdof     Degrees of freedom per node 
c                                   DOF reference numbers are:
c         UX  = 1, UY  = 2, UZ  = 3, ROTX= 4, ROTY= 5, ROTZ= 6, AX  = 7, AY  = 8
c         AZ  = 9, VX  =10, VY  =11, VZ  =12  ****** 13-18 are spares **********
c         ****************  PRES=19, TEMP=20, VOLT=21, MAG =22, ENKE=23, ENDS=24
c         EMF =25, CURR=26  ********* 27-32 are spares *************************
c                                    (curdof(i),i=1,numdof)

c  BAC      i       1    lenbac     Nodal equivalence table. This table equates
c                                   the number used for storage to the actual 
c                                   node number 
c                                    (Back(i),i=1,lenbac)

c  ELM      i       1     nume      Element equivalence table. The ANSYS program
c                                   stores all element data in the numerical 
c                                   order that the SOLUTION processor solves the
c                                   elements.  This table equates the order 
c                                   number used to the actual element number
c                                    (Order(i),i=1,nume)

c  GBK      i       1  Glblenbac    Global nodal equivalence table.  This
c                                   table equates the number used for storage
c                                   to the actual node number.  Only written
c                                   by the master process in Distributed Ansys
c                                    (GlbBack(i),i=1,Glblenbac)

c  FST      i       1     lenu      First element at a DOF table. This record 
c                                   signifies the first element encountered at a
c                                   particular DOF.
c                                    (First(i),i=1,lenu)

c  LST      i       1     lenu      Last element at a DOF table. This record 
c                                   signifies the last element encountered at a
c                                   particular DOF.
c                                    (Last(i),i=1,lenu)

c  BIT      i       1     lenu      Bits set at a DOF table. This record 
c                                   has bits for constraints, forces, etc.
c                                    (DofBits(i),i=1,lenu) (added at 10.0)

c  IDX      i       1     2*nume    Element index table. This record specifies 
c                                   the file location for the beginning of the
c                                   data for each element.
c                                    (index(i),i=1,nume) Low part of pointer
c                                    (index(i),i=1,nume) High part of pointer


c  The records at the end of the file store element information and get written
c  as a set for each element(nume sets of these records will appear on the file
c  at this point) ptrEHD indicates the beginning of the element data.

c  If substructure matrices are written to the EMAT file, they are written in a
c  different format than is shown here. This alternate format is not documented
c  at this time, as it is likely to change in the future.


c  EHD      i       1      10       Element matrix header
c                                   
c                                    stkey,   mkey,   dkey,  sskey, akey,
c                                    nrkey,   ikey,      0,      0, nmrow

c                                  each item in header is described below:

c                                   stkey  - stiffness matrix key 
c                                            = 0 - matrix not present
c                                            = 1 - matrix present
c                                   mkey   - mass matrix key 
c                                            = 0 - matirx not present
c                                            = 1 - matrix present
c                                   dkey   - damping matrix key 
c                                            = 0 - matrix not present
c                                            = 1 - matrix present
c                                   sskey  - stress stiffening matrix key 
c                                            = 0 - matrix not present
c                                            = 1 - matrix present
c                                   akey   - applied load vector key 
c                                            = 0 - vector not used
c                                            = 1 - vector used
c                                   nrkey  - newton-raphson(restoring) load 
c                                            vector key (for nonlinear analyses)
c                                            = 0 - vector not used
c                                            = 1 - vector used
c                                   ikey   - imaginary load vector key 
c                                              (for complex analyses)
c                                            = 0 - vector not used
c                                            = 1 - vector used 
c                                   0      - position not used
c                                   0      - position not used
c                                   nmrow  - numbers/columns in matrices. If the
c                                            number is negative, the matrices 
c                                            will be written in lower triangular
c                                            form.

c  ---      i       1    nmrow      DOF index table. This record specifies the 
c                                   DOF locations of this element matrix in 
c                                   relation to the global matrix. The index is
c                                   calculated as (N-1)*NUMDOF+DOF, where N is 
c                                   the position number of the node in the nodal
c                                   equivalence table and DOF is the DOF 
c                                   reference number given above

c  ---     dp    varies  varies     Element matrices. This record is repeated 
c                                   for each stiffness, mass, damping, and 
c                                   stress stiffening matrix. If the matrix is 
c                                   diagonal, the length of the records will be
c                                   nmrow.  If the matrix is unsymmetric, the 
c                                   length of the records will be nmrow*nmrow. 
c                                   If the matrix is symmetric, only the lower 
c                                   triangular terms are written and the length
c                                   of the records will be (nmrow)*(nmrow+1)/2.

c  ---     dp       1    2*nmrow    Element force vectors. This record contains
c                                   both the applied force vector and the 
c                                   (restoring or imaginary) load vector.
c
c 
c      *************** Internal CE information ***********************
c      The following records repeat numCE times... one for each internal
c      CE created during solution... these are stored here for the 
c      usage of a prestressed modal analysis such as the linear perturbation analysis
c
c  CE      i       3     numCE      First part is the CE number, the second part is
c                                   the number of terms in this internal CE, and
c                                   the third part is the external element number
c                                   of the element that created this internal CE
c
c  ---     i     nTerms  numCE      integer info (list of node*32 + dof)
c                                                     
c  ---    dp     nTerms  numCE      dp info (list of coefficients including constant term)
c
c
c   kygst        global stiffness matrix calculate key
c   kygm         global mass matrix calculate key
c   kygd         global damping matrix calculate key
c   kygss        global stress stiffening matrix calculate key
c   kygaf        global applied force matrix calculate key
c   kygrf        global restoring force matrix calculate key

"""
import numpy as np

from pyansys.common import read_table, parse_header

EMAT_HEADER_KEYS = ['fun02', 'nume', 'numdof', 'lenu', 'lenbac',
                    'maxn', 'nlgeEMA', 'sstEMAT', 'nodref', 'lumpm',
                    'kygst', 'kygm', 'kycd', 'kygss', 'kygaf',
                    'kygrf', '0', 'Glblenbac', 'ptrGBkl', 'ptrGBkh',
                    'ptrElmh', 'ptrFSTh', 'ptrLSTh', 'ptrBITh',
                    'ptrEHDh', 'ptrIDXh', 'numCE', 'maxLeng',
                    'ptrCEl', 'ptrCEh', 'ptrDOF', 'ptrBAC', 'ptrElml',
                    'ptrFSTl', 'ptrLSTl', 'ptrBITl', 'ptrEHDl',
                    'ptrIDXl', 'ptrendH', 'ptrendL']

ELEMENT_HEADER_KEYS = ['stkey', 'mkey', 'dkey', 'sskey', 'akey',
                       'nrkey', 'ikey', '_', '_', 'nmrow']


class EmatFile(object):
    """Enables pythonic access for an ANSYS element matrix file.

    Parameters
    ----------
    filename : str
        File to open.  Usually ends in .emat

    Examples
    --------
    >>> import pyansys
    >>> emat_file = pyansys.read_binary('file.emat')
    """

    def __init__(self, filename):
        self._element_matrices_index_table = None
        self._element_equivalence_table = None
        self._neqv = None
        self._nnum = None
        self._eeqv = None
        self._enum = None
        self.filename = filename
        self.read_header()

    def read_header(self):
        """Read standard emat file header"""
        with open(self.filename, 'rb') as f:
            f.seek(103*4)
            self.header = parse_header(read_table(f), EMAT_HEADER_KEYS)

    def read_element_matrix_header(self, f_index):
        """Read element matrix header

        Parameters
        ----------
        f_indes : int
            Fortran index to the start of the element matrix header.

        Notes
        -----
        stkey - stiffness matrix key 
            0 - matrix not present
            1 - matrix present

        mkey - mass matrix key 
            0 - matirx not present
            1 - matrix present

        dkey - damping matrix key 
            0 - matrix not present
            1 - matrix present

        sskey - stress stiffening matrix key 
            0 - matrix not present
            1 - matrix present

        akey - applied load vector key 
            0 - vector not used
            1 - vector used

        nrkey - newton-raphson(restoring) load 
            0 - vector not used
            1 - vector used

        ikey - imaginary load vector key (for complex analyses)
            0 - vector not used
            1 - vector used 

        nmrow - numbers/columns in matrices. 
            If the number is negative, the matrices will be written in
            lower triangular form.

        """
        with open(self.filename, 'rb') as f:
            f.seek(4*f_index)
            return parse_header(read_table(f), ELEMENT_HEADER_KEYS)

    @property
    def element_matrices_index_table(self):
        """Return element matrices index table"""
        if self._element_matrices_index_table is None:
            with open(self.filename, 'rb') as f:
                f.seek(self.header['ptrIDX']*4)
                self._element_matrices_index_table = read_table(f)
        return self._element_matrices_index_table


    # def element_matricies(self, index):
    #     """Element matrices

    #     This record is repeated for each stiffness, mass, damping, and
    #     stress stiffening matrix. If the matrix is diagonal, the
    #     length of the records will benmrow.  If the matrix is
    #     unsymmetric, the length of the records will be nmrow*nmrow. If
    #     the matrix is symmetric, only the lower triangular terms are
    #     written and the lengthof the records will be
    #     (nmrow)*(nmrow+1)/2.
    #     """
    #     self.element_matrices_index_table[index]

    @property
    def neqv(self):
        """Nodal equivalence table. This table equates the number used
        for storage to the actual node number.
        """
        if self._neqv is None:
            with open(self.filename, 'rb') as f:
                f.seek(self.header['ptrBAC']*4)
                self._neqv = read_table(f)
        return self._neqv

    @property
    def nnum(self):
        """Sorted ANSYS node numbers"""
        if self._nnum is None:
            self._nnum = np.sort(self.neqv)
        return self._nnum

    @property
    def eeqv(self):
        """Element equivalence table.  This table equates the number
        used for storage to the actual element number.

        Notes
        -----
        The ANSYS program stores all element data in the numerical
        order that the SOLUTION processor solves the elements.  This
        table equates the order number used to the actual element.
        """
        if self._eeqv is None:
            with open(self.filename, 'rb') as f:
                f.seek(self.header['ptrElm']*4)
                self._eeqv = read_table(f)
        return self._eeqv

    @property
    def enum(self):
        """Sorted ANSYS element numbers"""
        if self._enum is None:
            self._enum = np.sort(self.eeqv)
        return self._enum

    def read_element(self, index, stress=True, mass=True,
                     damping=True, stress_stiff=True,
                     applied_force=True, newton_raphson=True,
                     imaginary_load=True):
        """Read element by index

        Parameters
        ----------
        index : int
            Element index.  This is not the element number.  Reference
            the element equivalency table for the actual element
            number.

        stress : bool, optional
            Return the stress matrix entries if available.

        mass : bool, optional
            Return the mass matrix entries if available.

        damping : bool, optional
            Return the damping matrix entries if available.

        stress_stiff : bool, optional
            Return the stress stiffening entries if available.

        newton_raphson : bool, optional
            Return the newton raphson load entries if available.

        applied_force : bool, optional
            Return the applied load vector if available.

        imaginary_load : bool, optional
            Return the imaginary load vector if available.

        Returns
        -------
        dof_idx : np.ndarray
            DOF index table. This record specifies the DOF locations
            of this element matrix in relation to the global
            matrix. The index is calculated as (N-1)*NUMDOF+DOF, where
            N is the position number of the node in the nodal
            equivalence table and DOF is the DOF reference number
            given above

        element_data : dict
            Dictionary containing the following entries for each of
            the corresponding inputs when the item is True.
            - 'stress' : stress matrix entries
            - 'mass' : mass matrix entries
            - 'damping' : damping matrix entries
            - 'stress_stiff' : stress stiffening matrix entries
            - 'newton_raphson' : newton rapson load vector
            - 'applied_force' : applied force vector
            - 'imaginary_load' : imaginary load vector

        Notes
        -----
        If the matrix is diagonal, the length of the records will be
        nmrow.  If the matrix is unsymmetric, the length of the
        records will be nmrow*nmrow. If the matrix is symmetric, only
        the lower triangular terms are written and the length of the
        records will be ``(nmrow)*(nmrow+1)/2``

        Records are written relative to the dof_idx.  The index is
        calculated as (N-1)*NUMDOF+DOF, where N is the position number
        of the node in the nodal equivalence table and DOF is the DOF
        reference number given by ``dof_idx``.
        """
        element_data = {}
        with open(self.filename, 'rb') as fobj:
            # seek to the position of the element table in the file
            fobj.seek(self.element_matrices_index_table[index]*4)

            # matrix header
            element_header = parse_header(read_table(fobj), ELEMENT_HEADER_KEYS)
            nmrow = abs(element_header['nmrow'])
            lower_tri = element_header['nmrow'] < 0
            if lower_tri:
                nread = nmrow*(nmrow + 1)//2
            else:
                nread = nmrow*nmrow

            # Read DOF index table. This record specifies the DOF locations
            # of this element matrix in relation to the global
            # matrix. The index is calculated as (N-1)*NUMDOF+DOF,
            # where N is the position number of the node in the nodal
            # equivalence table and DOF is the DOF reference number
            # given above
            dof_idx = read_table(fobj) - 1  # adj one based indexing

            # read stress matrix
            if element_header['stkey']:  # if entry even exists
                if stress:
                    stress_entries = read_table(fobj, 'd', nread)
                    element_data['stress'] = stress_entries
                else:  # skip
                    fobj.seek(nread*8 + 12, 1)

            # read mass matrix
            if element_header['mkey']:  # if entry even exists
                if mass:
                    mass_entries = read_table(fobj, 'd', nread)
                    element_data['mass'] = mass_entries
                else:  # skip
                    fobj.seek(nread*8 + 12, 1)

            if element_header['dkey']:
                if damping:
                    damping_entries = read_table(fobj, 'd', nread)
                    element_data['damping'] = damping_entries
                else:  # skip
                    fobj.seek(nread*8 + 12, 1)

            if element_header['sskey']:
                if stress_stiff:
                    stress_stiff_entries = read_table(fobj, 'd', nread)
                    element_data['stress_stiff'] = stress_stiff_entries
                else:  # skip
                    fobj.seek(nread*8 + 12, 1)

            if element_header['akey']:
                if applied_force:
                    applied_force_vector = read_table(fobj, 'd', nmrow)
                    element_data['applied_force'] = applied_force_vector
                else:  # skip
                    fobj.seek(nmrow*8 + 12, 1)

            if element_header['nrkey']:
                if newton_raphson:
                    newton_raphson_vector = read_table(fobj, 'd', nmrow)
                    element_data['newton_raphson'] = newton_raphson_vector
                else:  # skip
                    fobj.seek(nmrow*8 + 12, 1)

            if element_header['ikey']:
                if imaginary_load:
                    imaginary_load_vector = read_table(fobj, 'd', nmrow)
                    element_data['imaginary_load'] = imaginary_load_vector
                else:  # skip
                    fobj.seek(nmrow*8 + 12, 1)

        return dof_idx, element_data

    @property
    def n_dof(self):
        """Number of dofs per node"""
        return self.header['numdof']

    @property
    def n_nodes(self):
        """Number of nodes in the file"""
        return self.header['lenbac']

    @property
    def n_elements(self):
        """Number of elements in the file"""
        return self.header['nume']

    def global_applied_force(self):
        """Returns the applied force for each node.

        Returns
        -------
        applied_force : np.ndarray
            Applied force with size (n_nodes, n_dof).  Result is
            sorted to correspond with the sorted global nodes array.
        """
        applied_force = np.zeros(self.n_nodes*self.n_dof)
        ncount = np.zeros(self.n_nodes*self.n_dof, np.int16)

        for i in range(self.n_elements):
            dof_idx, element_data = self.read_element(i, stress=False,
                                                       mass=False, damping=False)
            if 'applied_force' in element_data:
                applied_force[dof_idx] += element_data['applied_force']
                ncount[dof_idx] += 1

        # take the mean of each hit
        mask = ncount > 0
        applied_force[mask] /= ncount[mask]

        # resort applied force to correspond to sorted nodes
        dof_eqv = np.empty((self.n_nodes, self.n_dof), np.int32)
        dof_eqv[:] = self.neqv.reshape(-1, 1)*3
        dof_eqv[:, 1] += 1
        dof_eqv[:, 2] += 2

        s_idx = np.argsort(dof_eqv.ravel())
        applied_force = applied_force[s_idx].reshape((-1, 3))

        return applied_force



"""Module to control interaction with an ANSYS shell instance.

Used when launching Mapdl via pexpect on Linux when <= 17.0
"""
import pexpect
import time
import re

from pyansys.misc import kill_process
from pyansys.mapdl import _MapdlCore
from pyansys.errors import MapdlExitedError

ready_items = [
    rb'BEGIN:',
    rb'PREP7:',
    rb'SOLU_LS[0-9]+:',
    rb'POST1:',
    rb'POST26:',
    rb'RUNSTAT:',
    rb'AUX2:',
    rb'AUX3:',
    rb'AUX12:',
    rb'AUX15:',
    # continue
    rb'YES,NO OR CONTINUOUS\)\=',
    rb'executed\?',
    # errors
    rb'SHOULD INPUT PROCESSING BE SUSPENDED\?',
    # prompts
    rb'ENTER FORMAT for',
]

CONTINUE_IDX = ready_items.index(rb'YES,NO OR CONTINUOUS\)\=')
WARNING_IDX = ready_items.index(rb'executed\?')
ERROR_IDX = ready_items.index(rb'SHOULD INPUT PROCESSING BE SUSPENDED\?')
PROMPT_IDX = ready_items.index(rb'ENTER FORMAT for')

nitems = len(ready_items)
expect_list = []
for item in ready_items:
    expect_list.append(re.compile(item))
ignored = re.compile(r'[\s\S]+'.join(['WARNING', 'command', 'ignored']))


def launch_pexpect(exec_file=None, run_location=None, jobname=None, nproc=None,
                   additional_switches='', start_timeout=60):
    """Launch MAPDL as a pexpect process.

    Limited to only a linux instance
    """
    command = '%s -j %s -np %d %s' % (exec_file, jobname, nproc,
                                      additional_switches)
    process = pexpect.spawn(command, cwd=run_location)
    process.delaybeforesend = None

    try:
        index = process.expect(['BEGIN:', 'CONTINUE'],
                               timeout=start_timeout)
    except:  # capture failure
        raise RuntimeError(process.before.decode('utf-8'))

    if index:  # received ... press enter to continue
        process.sendline('')
        process.expect('BEGIN:', timeout=start_timeout)

    return process


class MapdlConsole(_MapdlCore):
    """Control interaction with an ANSYS shell instance.

    Only works on Linux.
    """

    def __init__(self, loglevel='INFO', log_apdl='w', use_vtk=True,
                 **start_parm):
        """Opens an ANSYS process using pexpect"""
        self._auto_continue = True
        self._continue_on_error = False
        self._process = None
        self._launch(start_parm)
        super().__init__(loglevel=loglevel, use_vtk=use_vtk, log_apdl=log_apdl,
                         **start_parm)

    def _launch(self, start_parm):
        """Connect to MAPDL process using pexpect"""
        self._process = launch_pexpect(**start_parm)

    def _run(self, command):
        """Sends command and returns ANSYS's response"""
        self._reset_cache()

        if not self._process.isalive():
            raise MapdlExitedError('ANSYS exited')

        command = command.strip()
        if not command:
            raise ValueError('Cannot run empty command')

        if command[:4].lower() == '/out':
            items = command.split(',')
            if len(items) > 1:
                self._output = '.'.join(items[1:])
            else:
                self._output = ''

        # send the command
        self._log.debug('Sending command %s', command)
        self._process.sendline(command)

        # do not expect
        if '/MENU' in command:
            self._log.info('Enabling GUI')
            self._process.sendline(command)
            return

        full_response = ''
        while True:
            i = self._process.expect_list(expect_list, timeout=None)
            response = self._process.before.decode('utf-8')
            full_response += response
            if i >= CONTINUE_IDX and i < WARNING_IDX:  # continue
                self._log.debug('Continue: Response index %i.  Matched %s',
                                i, ready_items[i].decode('utf-8'))
                self._log.info(response + ready_items[i].decode('utf-8'))
                if self._auto_continue:
                    user_input = 'y'
                else:
                    user_input = input('Response: ')
                self._process.sendline(user_input)

            elif i >= WARNING_IDX and i < ERROR_IDX:  # warning
                self._log.debug('Prompt: Response index %i.  Matched %s',
                                i, ready_items[i].decode('utf-8'))
                self._log.warning(response + ready_items[i].decode('utf-8'))
                if self._auto_continue:
                    user_input = 'y'
                else:
                    user_input = input('Response: ')
                self._process.sendline(user_input)

            elif i >= ERROR_IDX and i < PROMPT_IDX:  # error
                self._log.debug('Error index %i.  Matched %s',
                                i, ready_items[i].decode('utf-8'))
                self._log.error(response)
                response += ready_items[i].decode('utf-8')
                raise Exception(response)

            elif i >= PROMPT_IDX:  # prompt
                self._log.debug('Prompt index %i.  Matched %s',
                                i, ready_items[i].decode('utf-8'))
                self._log.info(response + ready_items[i].decode('utf-8'))
                raise RuntimeError('User input expected.  '
                                   'Try using ``with mapdl.non_interactive``')
            else:  # continue item
                self._log.debug('continue index %i.  Matched %s',
                                i, ready_items[i].decode('utf-8'))
                break

        # return last response and all preceding responses
        return full_response

    def exit(self, close_log=True, timeout=3):
        """Exit MAPDL process.

        Parameters
        ----------
        timeout : float
            Maximum time to wait for MAPDL to exit.  Set to 0 or
            ``None`` to not wait until MAPDL stops.
        """
        self._log.debug('Exiting ANSYS')
        if self._process is not None:
            try:
                self._process.sendline('FINISH')
                self._process.sendline('EXIT')
            except:
                pass

        if close_log:
            self._close_apdl_log()

        self._exited = True

        # edge case: need to wait until process dies, otherwise future
        # commands might talk to a dead process...
        if timeout:
            tstart = time.time()
            while self._process.isalive():
                time.sleep(0.05)
                telap = tstart - time.time()
                if telap > timeout:
                    return 1

        return 0

    def kill(self):
        """ Forces ANSYS process to end and removes lock file """
        if self._process is not None:
            try:
                self.exit()
            except:
                kill_process(self._process.pid)
                self._log.debug('Killed process %d' % self._process.pid)



"""
Provides a basic interface to computing the minimum scaled Jacobian
cell quality from VTK unstructured grids.

"""
import numpy as np

import pyvista as pv
from pyansys._cellqual import cell_quality_float, cell_quality
from pyansys.misc import vtk_cell_info


def quality(grid):
    """Compute the minimum scaled Jacobian cell quality of an UnstructuredGrid.

    Negative values indicate invalid cells while positive values
    indicate valid cells.  Varies between -1 and 1.

    Examples
    --------
    >>> import pyansys
    >>> import pyvista as pv
    >>> x = np.arange(-10, 10, 5)
    >>> y = np.arange(-10, 10, 5)
    >>> z = np.arange(-10, 10, 5)
    >>> x, y, z = np.meshgrid(x, y, z)
    >>> grid = pv.StructuredGrid(x, y, z)
    >>> pyansys.quality(grid)
    array([1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1.,
           1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1.])
    """
    flip = False
    if isinstance(grid, pv.StructuredGrid):
        grid = grid.cast_to_unstructured_grid()
        flip = True
    elif not isinstance(grid, pv.UnstructuredGrid):
        grid = pv.wrap(grid)
        if not isinstance(grid, pv.UnstructuredGrid):
            raise TypeError('Input grid should be a pyvista or vtk UnstructuredGrid')

    celltypes = grid.celltypes
    points = grid.points
    cells, offset = vtk_cell_info(grid)
    if points.dtype == np.float64:
        qual = cell_quality(cells, offset, celltypes, points)
    elif points.dtype == np.float32:
        qual = cell_quality_float(cells, offset, celltypes, points)
    else:
        points = points.astype(np.float64)
        qual = cell_quality(cells, offset, celltypes, points)

    # set qual of null cells to 1
    qual[grid.celltypes == 0] = 1
    if flip:  # for strucutred grids
        return -qual
    return qual
